import pandas as pd


class Model(object):
    """Internal utility for generating the definition of multi-table and single-table datasets.

    Create the definition of a dataset containing one or more tables. The definition includes the name and description
    of the dataset and the name and description of each table, attribute, and metric within the dataset.

    Attributes:

    """

    # dictionary key names
    _KEY_TABLE_NAME = 'table_name'
    _KEY_DATA_FRAME = 'data_frame'
    _KEY_AS_ATTR = 'to_attribute'
    _KEY_AS_METR = 'to_metric'
    _KEY_UPDATE_POL = 'update_policy'
    _KEY_TABLES = [_KEY_TABLE_NAME, _KEY_DATA_FRAME, _KEY_AS_ATTR, _KEY_AS_METR]

    __MAX_DESC_LEN = 250

    _INVALID_COL_CHARS = [' ', '.']  # check for invalid characters in column names

    def __init__(self, tables, name, description=None, folder_id=None, ignore_special_chars=False):
        """Initializes Model with tables, a name, and an optional description.

        Args:
            tables (:obj:`list` of :obj:`dict`): List of one or more dictionaries with keys `table_name`, `data_frame`,
                and optionally `as_attribute` and `as_metric`. Note that `as_attribute` and `as_metric` should be used
                when the default Python data type (e.g. `int`) should be converted to an attribute instead of a metric
                in MicroStrategy.
            name (str): Name of the data set.
            description (str, optional): Description of the data set. Must be less than or equal to 250 characters.
            folder_id (str, optional): ID of the shared folder that the dataset should be created within. If `None`,
                defaults to the user's My Reports folder.

        """

        self.__ignore_special_chars = ignore_special_chars

        # check integrity of tables list
        self.__check_table_list(tables=tables)

        # check dataset name params
        self.__name = name
        self.__check_param_str(self.__name, msg="Dataset name should be a string.")
        self.__check_param_len(self.__name,
                               msg="Dataset name should be <= {} characters.".format(self.__MAX_DESC_LEN),
                               length=self.__MAX_DESC_LEN)

        # check dataset description params
        if description is None:
            self.__description = ""
        else:
            self.__description = description
            self.__check_param_str(self.__description, msg="Dataset description should be a string.")
            self.__check_param_len(self.__description,
                                   msg="Dataset description should be <= {} characters.".format(self.__MAX_DESC_LEN),
                                   length=self.__MAX_DESC_LEN)

        # check folder_id param
        if folder_id is None:
            self._folder_id = ""
        else:
            self._folder_id = folder_id

        # init lists to accumulate table, attr, metric definitions and model object
        self.__tables = []
        self.__attributes = []
        self.__metrics = []
        self.__model = None

        # build the model
        self.__build(tables=tables)

    def get_model(self):
        """Return the model object."""
        return self.__model

    def __build(self, tables):
        """Generates the data model by mapping attributes and metrics from list of tables."""

        for table in tables:

            # map column names and column types
            _col_names = self.__get_col_names(table[self._KEY_DATA_FRAME])
            _col_types = self.__get_col_types(table[self._KEY_DATA_FRAME])

            # map tables
            self.__add_table(name=table[self._KEY_TABLE_NAME], col_names=_col_names, col_types=_col_types)

            # map attributes and metrics
            for _name, _type in zip(_col_names, _col_types):

                if self.__is_metric(_type):
                    if self._KEY_AS_ATTR in table.keys() and _name in table[self._KEY_AS_ATTR]:
                        self.__add_attribute(_name, table[self._KEY_TABLE_NAME])
                    else:
                        self.__add_metric(_name, table[self._KEY_TABLE_NAME])

                else:
                    if self._KEY_AS_METR in table.keys() and _name in table[self._KEY_AS_METR]:
                        self.__add_metric(_name, table[self._KEY_TABLE_NAME])
                    else:
                        self.__add_attribute(_name, table[self._KEY_TABLE_NAME])

        # set model object
        self.__model = {"name": self.__name,
                        "description": self.__description,
                        "folderId": self._folder_id,
                        "tables": self.__tables,
                        "metrics": self.__metrics,
                        "attributes": self.__attributes}

    def __get_col_types(self, table):
        """Map column types from each column in the list of table."""
        return list(map(self.__map_data_type, list(table.dtypes.values)))

    def __add_metric(self, name, table_name):
        """Add a metric to a metric list instance."""
        self.__metrics.append({'name': name, 'expressions': [{'tableName': table_name, 'columnName': name}]})

    def __add_attribute(self, name, table_name):
        """Add an attribute to an attribute list instance."""
        self.__attributes.append({'name': name,
                                  'attributeForms': [{'category': 'ID',
                                                      'expressions': [{'tableName': table_name,
                                                                       'columnName': name}]}]})

    def __add_table(self, name, col_names, col_types):
        """Add a table to a table list instance."""
        self.__tables.append({'name': name,
                              'columnHeaders': [{'name': name, 'dataType': typ} for name, typ in
                                                zip(col_names, col_types)]})

    def __check_table_list(self, tables):
        """Check integrity of table list parameter."""

        # tables must be a list
        if not isinstance(tables, list):
            raise TypeError("Elements of tables must be a list of dicts.")

        # tables cannot be length 0
        if len(tables) is 0:
            raise ValueError("No tables have been added to the dataset.")

        # check integrity of each table passed to tables
        for table in tables:
            self.__check_table(table)

    def __check_table(self, table):
        """Check integrity of table parameter."""

        # force all list elements to be a dict with specific names
        msg = "Each table must be a dictionary with keys: '{}', '{}', '{},' and '{}'.".format(*self._KEY_TABLES)
        if not isinstance(table, dict):
            raise TypeError(msg)

        if not all(k in table.keys() for k in (self._KEY_TABLE_NAME, self._KEY_DATA_FRAME)):
            raise ValueError(msg)

        # check that the value of the data frame key is a pandas data frame
        if not isinstance(table[self._KEY_DATA_FRAME], pd.DataFrame):
            msg = "Pandas DataFrame must be passed as the value in the '{}' key.".format(self._KEY_DATA_FRAME)
            raise TypeError(msg)

        # check for presence of invalid characters in data frame column names
        if not self.__ignore_special_chars:
            if any([col for col in table[self._KEY_DATA_FRAME].columns for inv in self._INVALID_COL_CHARS if inv in col]):
                msg = "Column names cannot contain '{}' or '{}'".format(*self._INVALID_COL_CHARS)
                raise ValueError(msg)

    @staticmethod
    def __get_col_names(table):
        """Returns column names from a table as a list."""
        return list(table.columns)

    @staticmethod
    def __map_data_type(datatype):
        """Maps a Python data type to a MicroStrategy data type."""
        if datatype == 'object':
            return "STRING"
        elif datatype in ['int64', 'int32']:
            return "INTEGER"
        elif datatype in ['float64', 'float32']:
            return "DOUBLE"
        elif datatype == 'bool':
            return "BOOLEAN"
        elif datatype == 'datetime64[ns]':
            return 'DATETIME'

    @staticmethod
    def __check_param_len(param, msg, length):
        if len(param) >= length:
            raise ValueError(msg)
        else:
            return True

    @staticmethod
    def __check_param_str(param, msg):
        if not isinstance(param, str):
            raise TypeError(msg)
        else:
            return True

    @staticmethod
    def __is_metric(datatype):
        """Helper function for determining if the requested datatype is (by default) a metric or attribute."""
        if datatype in ["DOUBLE", "INTEGER"]:
            return True
        else:
            return False

    @property
    def name(self):
        return self.__name

    @property
    def description(self):
        return self.__description

    @property
    def folder_id(self):
        return self._folder_id

    @property
    def tables(self):
        return self.__tables

    @property
    def attributes(self):
        return self.__attributes

    @property
    def metrics(self):
        return self.__metrics



import pandas
import unittest
from unittest.mock import Mock, patch

from mstrio.report import Report


class TestReport(unittest.TestCase):

    def setUp(self):
        self.report_id = '1DC7D33611E9AE27F6B00080EFE52FBC'
        self.report_name = 'UnitTest'
        self.connection = "test_connection"
        self.__definition = {'id': self.report_id,
                            'name': self.report_name,
                             'result': {'definition': {'availableObjects': {'metrics': [{'name': 'Age',
                                'id': '089FB58611E9CA4D39700080EF15B5B9',
                                'type': 'Metric'},
                                {'name': 'Row Count - table1',
                                'id': '089DE7BA11E9CA4D085B0080EFC515B9',
                                'type': 'Metric'}],
                                'attributes': [{'name': 'Name',
                                'id': '089FC10C11E9CA4D39700080EF15B5B9',
                                'type': 'Attribute',
                                'forms': [{'id': '45C11FA478E745FEA08D781CEA190FE5',
                                    'name': 'ID',
                                    'dataType': 'Char',
                                    'baseFormCategory': 'ID',
                                    'baseFormType': 'Text'}]}]}}}}
        self.__instance = {'id': self.report_id,
                            'name': self.report_name,
                            'status': 1,
                            'instanceId': '49C2D26C11E9CB21237E0080EF1546B6',
                            'result': {'definition': {'metrics': [{'name': 'Age',
                                'id': '089FB58611E9CA4D39700080EF15B5B9',
                                'type': 'Metric',
                                'min': 18,
                                'max': 21,
                                'dataType': 'Integer',
                                'numberFormatting': {'category': 9, 'formatString': 'General'}}],
                            'attributes': [{'name': 'Name',
                                'id': '089FC10C11E9CA4D39700080EF15B5B9',
                                'type': 'Attribute',
                                'forms': [{'id': '45C11FA478E745FEA08D781CEA190FE5',
                                'name': 'ID',
                                'dataType': 'Char',
                                'baseFormCategory': 'ID',
                                'baseFormType': 'Text'}]}],
                            'thresholds': [],
                            'sorting': []},
                            'data': {'paging': {'total': 4,
                                'current': 2,
                                'offset': 0,
                                'limit': 2,
                                'prev': None,
                                'next': None},
                            'root': {'isPartial': True,
                                'children': [{'depth': 0,
                                'element': {'attributeIndex': 0,
                                'formValues': {'ID': 'jack'},
                                'name': 'jack',
                                'id': 'hjack;089FC10C11E9CA4D39700080EF15B5B9'},
                                'metrics': {'Age': {'rv': 18, 'fv': '18', 'mi': 0}}},
                                {'depth': 0,
                                'element': {'attributeIndex': 0,
                                'formValues': {'ID': 'krish'},
                                'name': 'krish',
                                'id': 'hkrish;089FC10C11E9CA4D39700080EF15B5B9'},
                                'metrics': {'Age': {'rv': 19, 'fv': '19', 'mi': 0}}}]}}}}
        self.__instance_id = {'id': self.report_id,
                            'name': self.report_name,
                            'status': 1,
                            'instanceId': '49C2D26C11E9CB21237E0080EF1546B6',
                            'result': {'definition': {'metrics': [{'name': 'Age',
                                'id': '089FB58611E9CA4D39700080EF15B5B9',
                                'type': 'Metric',
                                'min': 18,
                                'max': 21,
                                'dataType': 'Integer',
                                'numberFormatting': {'category': 9, 'formatString': 'General'}}],
                            'attributes': [{'name': 'Name',
                                'id': '089FC10C11E9CA4D39700080EF15B5B9',
                                'type': 'Attribute',
                                'forms': [{'id': '45C11FA478E745FEA08D781CEA190FE5',
                                'name': 'ID',
                                'dataType': 'Char',
                                'baseFormCategory': 'ID',
                                'baseFormType': 'Text'}]}],
                            'thresholds': [],
                            'sorting': []},
                            'data': {'paging': {'total': 4,
                                'current': 2,
                                'offset': 2,
                                'limit': 2,
                                'prev': None,
                                'next': None},
                            'root': {'isPartial': True,
                                'children': [{'depth': 0,
                                'element': {'attributeIndex': 0,
                                'formValues': {'ID': 'nick'},
                                'name': 'nick',
                                'id': 'hnick;089FC10C11E9CA4D39700080EF15B5B9'},
                                'metrics': {'Age': {'rv': 21, 'fv': '21', 'mi': 0}}},
                                {'depth': 0,
                                'element': {'attributeIndex': 0,
                                'formValues': {'ID': 'Tom'},
                                'name': 'Tom',
                                'id': 'hTom;089FC10C11E9CA4D39700080EF15B5B9'},
                                'metrics': {'Age': {'rv': 20, 'fv': '20', 'mi': 0}}}]}}}}

        self.__attr_elements = [{'id': '089FC10C11E9CA4D39700080EF15B5B9:jack', 'formValues': ['jack']},
                                {'id': '089FC10C11E9CA4D39700080EF15B5B9:krish', 'formValues': ['krish']},
                                {'id': '089FC10C11E9CA4D39700080EF15B5B9:nick', 'formValues': ['nick']},
                                {'id': '089FC10C11E9CA4D39700080EF15B5B9:Tom', 'formValues': ['Tom']}]
        self.__headers = {'x-mstr-total-count': '4'}
        self.__selected_attr = ['089FC10C11E9CA4D39700080EF15B5B9']
        self.__selected_metrs = ['089FB58611E9CA4D39700080EF15B5B9']
        self.__selected_elem = ['089FC10C11E9CA4D39700080EF15B5B9:Tom', '089FC10C11E9CA4D39700080EF15B5B9:jack']
        self.__dataframe = pandas.DataFrame({'Name':['jack', 'krish','nick','Tom'], 'Age':[18,19,21,20]})

    @patch('mstrio.api.reports.report_single_attribute_elements')
    @patch('mstrio.api.reports.report')
    def test_init_report(self, mock_definition, mock_attr_element):
        """Test that definition of the report is assigned properly when report is initialized."""

        mock_definition.return_value = Mock(ok=True)
        mock_definition.return_value.json.return_value = self.__definition
        mock_attr_element.return_value = Mock(headers=self.__headers)
        mock_attr_element.return_value.json.return_value = self.__attr_elements

        report = Report(connection=self.connection, report_id=self.report_id)

        self.assertTrue(mock_definition.called)
        self.assertTrue(mock_attr_element.called)
        
        self.assertEqual(report._connection, self.connection)
        self.assertEqual(report._report_id, self.report_id)
        self.assertEqual(report.name, self.report_name)

        self.assertEqual(report.attributes, [{'name': 'Name', 'id': '089FC10C11E9CA4D39700080EF15B5B9'}])
        self.assertEqual(report.metrics, [{'name': 'Age', 'id': '089FB58611E9CA4D39700080EF15B5B9'},
                                        {'name': 'Row Count - table1', 'id': '089DE7BA11E9CA4D085B0080EFC515B9'}])
        self.assertEqual(report.attr_elements, [{'attribute_name': 'Name',
                                        'attribute_id': '089FC10C11E9CA4D39700080EF15B5B9','elements': [
                                        {'id': '089FC10C11E9CA4D39700080EF15B5B9:jack','formValues': ['jack']},
                                        {'id': '089FC10C11E9CA4D39700080EF15B5B9:krish', 'formValues': ['krish']},
                                        {'id': '089FC10C11E9CA4D39700080EF15B5B9:nick', 'formValues': ['nick']},
                                        {'id': '089FC10C11E9CA4D39700080EF15B5B9:Tom', 'formValues': ['Tom']}]}])

        self.assertIsNone(report.selected_attributes)
        self.assertIsNone(report.selected_metrics)
        self.assertIsNone(report.selected_attr_elements)
        self.assertIsNone(report._dataframe)
        with self.assertWarns(Warning):
            report.dataframe

    @patch('mstrio.api.reports.report_single_attribute_elements')
    @patch('mstrio.api.reports.report')
    def test_apply_filters(self, mock_definition, mock_attr_element):
        """Test that selected objects are assigned properly when filter is applied."""

        mock_definition.return_value = Mock(ok=True)
        mock_definition.return_value.json.return_value = self.__definition
        mock_attr_element.return_value = Mock(ok=True, headers=self.__headers)
        mock_attr_element.return_value.json.return_value = self.__attr_elements

        report = Report(connection=self.connection, report_id=self.report_id)
        report.apply_filters(self.__selected_attr, self.__selected_metrs, self.__selected_elem)

        self.assertEqual(report.selected_attributes, self.__selected_attr)
        self.assertEqual(report.selected_metrics, self.__selected_metrs)
        self.assertEqual(report.selected_attr_elements, self.__selected_elem)

        report.clear_filters()
        report.apply_filters(attributes=[], metrics=[])
        self.assertEqual(report.selected_attributes, [])
        self.assertEqual(report.selected_metrics, [])

    @patch('mstrio.api.reports.report_single_attribute_elements')
    @patch('mstrio.api.reports.report')
    def test_clear_filters(self, mock_definition, mock_attr_element):
        """Test that selected objects are assigned with empty lists when filter is cleared."""

        mock_definition.return_value = Mock(ok=True)
        mock_definition.return_value.json.return_value = self.__definition
        mock_attr_element.return_value = Mock(ok=True, headers=self.__headers)
        mock_attr_element.return_value.json.return_value = self.__attr_elements

        report = Report(connection=self.connection, report_id=self.report_id)
        report.apply_filters(self.__selected_attr, self.__selected_metrs, self.__selected_elem)

        self.assertEqual(report.selected_attributes, self.__selected_attr)
        self.assertEqual(report.selected_metrics, self.__selected_metrs)
        self.assertEqual(report.selected_attr_elements, self.__selected_elem)

        report.clear_filters()

        self.assertIsNone(report.selected_attributes)
        self.assertIsNone(report.selected_metrics)
        self.assertIsNone(report.selected_attr_elements)

    @patch('mstrio.api.reports.report_instance_id')
    @patch('mstrio.api.reports.report_instance')
    @patch('mstrio.api.reports.report_single_attribute_elements')
    @patch('mstrio.api.reports.report')
    def test_to_dataframe(self, mock_definition, mock_attr_element, mock_instance, mock_instance_id):
        """Test that data is retrieved and parsed properly when to_dataframe() is called.
        Result should be saved to Report.dataframe property.
        """

        mock_definition.return_value = Mock(ok=True)
        mock_definition.return_value.json.return_value = self.__definition
        mock_attr_element.return_value = Mock(ok=True, headers=self.__headers)
        mock_attr_element.return_value.json.return_value = self.__attr_elements
        mock_instance.return_value = Mock(ok=True)        
        mock_instance.return_value.json.return_value = self.__instance
        mock_instance_id.return_value = Mock(ok=True)
        mock_instance_id.return_value.json.return_value = self.__instance_id

        report = Report(connection=self.connection, report_id=self.report_id)
        df = report.to_dataframe(limit=2)
        
        self.assertTrue(mock_instance.called)
        self.assertTrue(mock_instance_id.called)
        self.assertIsInstance(df, pandas.core.frame.DataFrame)
        self.assertIsInstance(report.dataframe, pandas.core.frame.DataFrame)
        self.assertTrue(df.equals(self.__dataframe))

    @patch('mstrio.api.reports.report_single_attribute_elements')
    @patch('mstrio.api.reports.report')
    def test_apply_filters_for_incorrect_assignments(self, mock_definition, mock_attr_element):
        """Test that incorrectly assigned selected objects are assigned properly when filter is applied."""

        mock_definition.return_value = Mock(ok=True)
        mock_definition.return_value.json.return_value = self.__definition
        mock_attr_element.return_value = Mock(ok=True, headers=self.__headers)
        mock_attr_element.return_value.json.return_value = self.__attr_elements

        report = Report(connection=self.connection, report_id=self.report_id)
        # attributes assigned selected_metrs, metrics assigned selected_elem and attr_elements assigned selected_attr
        report.apply_filters(attributes=self.__selected_metrs, metrics=self.__selected_elem,
                             attr_elements=self.__selected_attr)

        self.assertEqual(report.selected_attributes, self.__selected_attr)
        self.assertEqual(report.selected_metrics, self.__selected_metrs)
        self.assertEqual(report.selected_attr_elements, self.__selected_elem)

    @patch('mstrio.api.reports.report_single_attribute_elements')
    @patch('mstrio.api.reports.report')
    def test_apply_filters_no_list(self, mock_definition, mock_attr_element):
        """Test that selected objects passed as strings are assigned properly when filter is applied."""

        mock_definition.return_value = Mock(ok=True)
        mock_definition.return_value.json.return_value = self.__definition
        mock_attr_element.return_value = Mock(ok=True, headers=self.__headers)
        mock_attr_element.return_value.json.return_value = self.__attr_elements

        report = Report(connection=self.connection, report_id=self.report_id)
        report.apply_filters(attributes=self.__selected_attr[0], metrics=self.__selected_metrs[0],
                             attr_elements=self.__selected_elem[0])

        self.assertEqual(report.selected_attributes, self.__selected_attr)
        self.assertEqual(report.selected_metrics, self.__selected_metrs)
        self.assertEqual(report.selected_attr_elements, self.__selected_elem)

    @patch('mstrio.api.reports.report_single_attribute_elements')
    @patch('mstrio.api.reports.report')
    def test_apply_filters_invalid_elements(self, mock_definition, mock_attr_element):
        """Test that invalid id passed to a filter raises ValueError."""

        mock_definition.return_value = Mock(ok=True)
        mock_definition.return_value.json.return_value = self.__definition
        mock_attr_element.return_value = Mock(ok=True, headers=self.__headers)
        mock_attr_element.return_value.json.return_value = self.__attr_elements

        report = Report(connection=self.connection, report_id=self.report_id)
        self.assertRaises(ValueError, report.apply_filters, attributes='INV123456')

if __name__ == '__main__':
    unittest.main()



# TODO (srigney): Create method to select project, call it 'select_project'
import warnings
import pandas as pd

from mstrio.dataset import Dataset
from mstrio.api import projects, cubes, reports, authentication, misc
from mstrio.utils.parsejson import parsejson
from packaging import version


class Connection(object):
    """Connect to and interact with the MicroStrategy environment.

    Creates a connection object which is used in subsequent requests and manages the user's connection
    with the MicroStrategy REST and Intelligence Servers.

    # import
    from mstrio import microstrategy

    # connect to the environment and chosen project
    conn = microstrategy.Connection(base_url="https://demo.microstrategy.com/MicroStrategyLibrary/api",
                                    username="username",
                                    password="password",
                                    project_name="MicroStrategy Tutorial")
    conn.connect()

    # disconnect
    conn.close()

    Attributes:
        base_url: URL of the MicroStrategy REST API server.
        username: Username.
        password: Password.
        project_name: Name of the connected MicroStrategy Project.
        project_id: Id of the connected MicroStrategy Project.
        login_mode: Authentication mode. Standard = 1 (default) or LDAP = 16.
        ssl_verify: If True (default), verifies the server's SSL certificates with each request.
    """

    auth_token = None
    cookies = None
    project_id = None
    application_code = 64
    __VRCH ="11.1.0400"

    def __init__(self, base_url=None, username=None, password=None, project_name=None, project_id=None,
                 login_mode=1, ssl_verify=True):
        """
        Establishes a connection with MicroStrategy REST API

        Args:
            base_url (str, optional): URL of the MicroStrategy REST API server. Typically of the form
            https://<mstr_env>.com/MicroStrategyLibrary/api
            username (str, optional): Username
            password (str, optional): Password
            project_name (str, optional): Name of the project you intend to connect to.Case-sensitive.
                            You should provide either Project ID or Project Name.
            project_id (str, optional): Id of the project you intend to connect to. Case-sensitive.
                            You should provide either Project ID or Project Name.
            login_mode (int, optional): Specifies the authentication mode to use. Supported authentication modes are: 
                            Standard (1) (default) or LDAP (16)
            ssl_verify (bool, optional): If True (default), verifies the server's SSL certificates with each request
        """

        self.base_url = base_url
        self.username = username
        self.password = password
        if project_id is not None and project_name is not None:
            raise ValueError("Provide either Project ID or Project Name.")
        else:
            self.project_id = project_id
            self.project_name = project_name
        self.login_mode = login_mode
        self.ssl_verify = ssl_verify
        self.__web_version = None
        self.__iserver_version = None

    def connect(self):
        """Creates a connection"""
        if self.__check_version():
            response = authentication.login(connection=self)

            if not response.ok:
                msg = 'Authentication error. Check user credentials or REST API URL and try again.'
                self.__response_handler(response=response, msg=msg)
            else:
                self.auth_token, self.cookies = response.headers['X-MSTR-AuthToken'], dict(response.cookies)

                # Fetch the list of projects the user has access to
                response = projects.projects(connection=self)

                if not response.ok:
                    msg = "Error connecting to project '{}'. Check project name and try again.".format(self.project_name)
                    self.__response_handler(response=response, msg=msg)
                else:
                    # Find which project ID matches the project name provided
                    _projects = response.json()

                    if self.project_name is not None:
                        # Find which project ID matches the project name provided
                        for _project in _projects:
                            if _project['name'] == self.project_name:  # Enter the desired project name here
                                self.project_id = _project['id']

                    else:
                        # Find which project name matches the project ID provided
                        for _project in _projects:
                            if _project['id'] == self.project_id:  # Enter the desired project id here
                                self.project_name = _project['name']
        else:
            print("""
            This version of mstrio is only supported on MicroStrategy 11.1.0400 or higher.
            Current Intelligence Server version: {}
            Current MicroStrategy Web version: {}
            """.format(self.__iserver_version,self.__web_version))
            raise VersionException('MicroStrategy Version not supported.')

    def close(self):
        """Closes a connection with MicroStrategy REST API"""
        authentication.logout(connection=self)

    def __check_version(self):
        """Checks version of I-Server and MicroStrategy Web"""
        response = misc.server_status(self)
        if not response.ok:
            msg = 'Failed to check server status'
            self.__response_handler(response=response, msg=msg)
        else:
            json_response = response.json()
            self.__iserver_version = json_response["iServerVersion"][:9]
            self.__web_version = json_response["webVersion"][:9]
            
            iserver_version_ok = version.parse(self.__iserver_version) >= version.parse(self.__VRCH)
            web_version_ok = version.parse(self.__web_version) >= version.parse(self.__VRCH)
            
            return iserver_version_ok and web_version_ok

    def get_report(self, report_id, offset=0, limit=1000):
        """
        Extracts the contents of a report into a Pandas Data Frame

        :param report_id: Unique ID of the report you wish to extract information from
        :param offset: (optional) To extract all data from the report, use 0 (default)
        :param limit: (optional) Used to control data extract behavior on datasets with a large
        number of rows. The default is 1000. As an example, if the dataset has 50,000 rows,
        get_report() will incrementally extract all 50,000 rows in 1,000 row chunks. Depending
        on system resources, a higher limit (e.g. 10,000) may reduce the total time
        required to extract the entire dataset
        :return: Pandas Data Frame containing the report contents
        """

        # warning for future deprecation / replacement by Report class
        warnings.warn(
            "This method will be deprecated. The Report constructor is preferred.",
            DeprecationWarning
        )

        response = reports.report_instance(connection=self,
                                           report_id=report_id,
                                           offset=offset,
                                           limit=limit)

        if not response.ok:
            msg = "Error getting report contents."
            self.__response_handler(response=response, msg=msg)
        else:
            json_response = response.json()
            instance_id = json_response['instanceId']

            # Gets the pagination totals from the response object
            pagination = json_response['result']['data']['paging']

            # If there are more rows to fetch, fetch them
            if pagination['current'] != pagination['total']:

                # initialize a list to capture slices from each query, and append the first request's result to the list
                table_data = [parsejson(response=json_response)]

                # Fetch add'l rows from this object instance from the intelligence server
                for _offset in range(limit, pagination['total'], limit):
                    response = reports.report_instance_id(connection=self,
                                                          report_id=report_id,
                                                          instance_id=instance_id,
                                                          offset=_offset,
                                                          limit=limit)
                    table_data.append(parsejson(response=response.json()))
                return pd.concat(table_data)
            else:
                return parsejson(response=json_response)

    def get_cube(self, cube_id, offset=0, limit=1000):
        """
        Extracts the contents of a cube into a Pandas Data Frame

        :param cube_id: Unique ID of the cube you wish to extract information from.
        :param offset: (optional) To extract all data from the report, use 0 (default)
        :param limit: (optional) Used to control data extract behavior on datasets with a large
        number of rows. The default is 1000. As an example, if the dataset has 50,000 rows,
        get_cube() will incrementally extract all 50,000 rows in 1,000 row chunks. Depending
        on system resources, a higher limit (e.g. 10,000) may reduce the total time
        required to extract the entire dataset
        :return: Pandas Data Frame containing the cube contents
        """

        # warning for future deprecation / replacement by Cube class
        warnings.warn(
            "This method will be deprecated. The Cube constructor is preferred and supports multi-table data.",
            DeprecationWarning
        )

        response = cubes.cube_instance(connection=self, cube_id=cube_id, offset=offset, limit=limit)

        if not response.ok:
            msg = "Error getting cube contents."
            self.__response_handler(response=response, msg=msg)
        else:
            json_response = response.json()
            instance_id = json_response['instanceId']

            # Gets the pagination totals from the response object
            pagination = json_response['result']['data']['paging']

            # If there are more rows to fetch, fetch them
            if pagination['current'] != pagination['total']:

                # initialize a list to capture slices from each query, and append the first request's result to the list
                table_data = [parsejson(response=json_response)]

                # Fetch add'l rows from this object instance from the intelligence server
                for _offset in range(limit, pagination['total'], limit):
                    response = cubes.cube_instance_id(connection=self,
                                                      cube_id=cube_id,
                                                      instance_id=instance_id,
                                                      offset=_offset,
                                                      limit=limit)
                    table_data.append(parsejson(response=response.json()))
                return pd.concat(table_data)
            else:
                return parsejson(response=json_response)

    def create_dataset(self, data_frame, dataset_name, table_name, to_metric=None, to_attribute=None, folder_id=None):
        """
        Create an in-memory MicroStrategy dataset from a Pandas Data Frame

        :param data_frame: A Pandas Data Frame from which an in-memory dataset will be created
        :param dataset_name: Name of the in-memory dataset
        :param table_name: Name of the table to create within the dataset
        :param to_metric: (optional) A vector of column names from the Data.Frame to format as metrics
        in the dataset. By default, numeric types are formatted as metrics while character and date types are formatted
        as attributes. For example, a column of integer-like strings ("1", "2", "3") would
        appear as an attribute in the newly created dataset. If the intent is to format this data as a metric, provide
        the corresponding column name as \code{to_metric=c('myStringIntegers')}
        :param to_attribute: (optional) Logical opposite of to_metric. Helpful for formatting an integer-based row
        identifier as a primary key in the dataset
        :param folder_id: (optional) ID of the shared folder that the dataset should be created within. If `None`,
            defaults to the user's My Reports folder.
        :return: Unique identifiers of the dataset and table within the newly created dataset. Required for
        update_dataset()
        """

        # warning for future deprecation / replacement by Datasets class
        warnings.warn(
            "This method will be deprecated. The Dataset constructor is preferred and supports multi-table data.",
            DeprecationWarning
        )

        # Replace any leading/trailing whitespace in df names, replace '.' with '_'
        _df = data_frame.copy()
        _df.columns = _df.columns.str.replace(".", "_")
        _df.columns = _df.columns.str.strip()

        if folder_id is None:
            folder_id = ""
        else:
            folder_id = folder_id

        # create dataset instance
        ds = Dataset(connection=self, name=dataset_name)

        # add table to the dataset
        ds.add_table(name=table_name, data_frame=_df, update_policy='add', to_metric=to_metric, to_attribute=to_attribute)

        # publish the dataset
        ds.create(folder_id=folder_id)

        return ds.dataset_id

    def update_dataset(self, data_frame, dataset_id, table_name, update_policy):

        """
        Update a previously created dataset with an Pandas Data Frame

        :param data_frame: Pandas Data Frame to use to update an in-memory dataset
        :param dataset_id: Identifier of the dataset to update, provided by create_dataset()
        :param table_name: Name of the table to update within the dataset
        :param update_policy: Update operation to perform. One of 'add' (inserts new, unique rows), 'update'
        (updates data in existing rows and columns), 'upsert' (updates existing
        data and inserts new rows), 'replace' (similar to truncate and load, replaces the existing data with new data)
        """

        # warning for future deprecation / replacement by Datasets class
        warnings.warn(
            "This method will be deprecated. The Dataset constructor is preferred and supports multi-table data.",
            DeprecationWarning
        )

        # Replace any leading/trailing whitespace in df names, replace '.' with '_'
        _df = data_frame.copy()
        _df.columns = _df.columns.str.replace(".", "_")
        _df.columns = _df.columns.str.strip()

        # create dataset instance, add table, then publish the updates to the dataset
        ds = Dataset(connection=self, dataset_id=dataset_id)
        ds.add_table(name=table_name, data_frame=_df, update_policy=update_policy)
        ds.update()
        ds.publish()

    def renew(self):
        """Checks if the session is still alive. If so, renews the session and extends session expiration."""
        response = authentication.sessions(connection=self)

        if not response.ok:
            msg = "Session expired. Please reconnect to MicroStrategy."
            self.__response_handler(response=response, msg=msg)
        else:
            print("MicroStrategy connection was renewed.")

    @staticmethod
    def __response_handler(response, msg):
        """Generic error message handler for transactions against datasets.

        Args:
            response: Response object returned by HTTP request.
            msg (str): Message to print in addition to any server-generated error message(s).

        """

        if response.status_code == 204:
            warnings.warn(
                '204 No Content: The server successfully processed the request and is not returning any content.')
        else:
            res = response.json()
            print("I-Server Error %s, %s" % (res['code'], res['message']))
            response.raise_for_status()

class VersionException(Exception):
    pass



import unittest
import pandas as pd
from mstrio.utils.model import Model


def make_df():
    # this is the base data frame used for testing
    raw_data = {"id_int": [1, 2, 3, 4, 5],
                "id_str": ["1", "2", "3", "4", "5"],
                "first_name": ["Jason", "Molly", "Tina", "Jake", "Amy"],
                "last_name": ["Miller", "Jacobson", "Turner", "Milner", "Cooze"],
                "age": [42, 52, 36, 24, 73],
                "weight": [100.22, 210.2, 175.1, 155.9, 199.9],
                "state": ["VA", "NC", "WY", "CA", "CA"],
                "salary": [50000, 100000, 75000, 85000, 250000]}
    df = pd.DataFrame(raw_data, columns=["id_int", "id_str", "first_name", "last_name",
                                         "age", "weight", "state", "salary"])
    return df


class TestModel(unittest.TestCase):

    def test_model_structure(self):
        df = make_df()
        tables = [{"table_name": "employee",
                   "data_frame": df}]

        DATASETNAME = "Employees"
        DESCRIPTION = "Table of employee data."

        model = Model(tables=tables, name=DATASETNAME, description=DESCRIPTION)
        model_dict = model.get_model()

        # check that model is a dictionary
        self.assertIsInstance(model_dict, dict)

        # check that dictionary keys are as expected
        self.assertCountEqual(list(model_dict.keys()), ["name", "description", "folderId",
                                                         "tables", "metrics", "attributes"])

        # check that these elements of the dictionary are lists
        self.assertIsInstance(model.tables, list)
        self.assertIsInstance(model.metrics, list)
        self.assertIsInstance(model.attributes, list)

        # check that name and description match
        self.assertEqual(model.name, DATASETNAME)
        self.assertEqual(model.description, DESCRIPTION)

        # check that column names are unaffected and match the original data frame
        ch = [i['name'] for i in model.tables[0]['columnHeaders']]
        self.assertCountEqual(list(df.columns), ch)

    def test_single_table(self):
        df = make_df()
        tables = [{"table_name": "employee",
                   "data_frame": df}]

        DATASETNAME = "Employees"
        DESCRIPTION = "Table of employee data."

        model = Model(tables=tables, name=DATASETNAME, description=DESCRIPTION)

        self.assertEqual(len(model.tables), 1)
        self.assertEqual(len(model.attributes), 4)
        self.assertEqual(len(model.metrics), 4)

    def test_multi_table(self):
        df = make_df()
        tables = [{"table_name": "employee1",
                   "data_frame": df},
                  {"table_name": "employee2",
                   "data_frame": df}]

        DATASETNAME = "Employees"
        DESCRIPTION = "Table of employee data."

        model = Model(tables=tables, name=DATASETNAME, description=DESCRIPTION)

        self.assertEqual(len(model.tables), 2)
        self.assertEqual(len(model.attributes), 8)
        self.assertEqual(len(model.metrics), 8)

    def test_metric_override(self):
        df = make_df()
        tables = [{"table_name": "employee",
                   "data_frame": df,
                   "to_metric": ["id_str"]}]

        DATASETNAME = "Employees"
        DESCRIPTION = "Table of employee data."

        model = Model(tables=tables, name=DATASETNAME, description=DESCRIPTION)

        self.assertEqual(len(model.attributes), 3)
        self.assertEqual(len(model.metrics), 5)

    def test_attribute_override(self):
        df = make_df()
        tables = [{"table_name": "employee1",
                   "data_frame": df,
                   "to_attribute": ["id_int"]},
                  {"table_name": "employee2",
                   "data_frame": df,
                   "to_attribute": ["id_int", "salary", "age"]}]

        DATASETNAME = "Employees"
        DESCRIPTION = "Table of employee data."

        model = Model(tables=tables, name=DATASETNAME, description=DESCRIPTION)

        self.assertEqual(len(model.attributes), 12)
        self.assertEqual(len(model.metrics), 4)

    def test_both_override(self):
        df = make_df()
        tables = [{"table_name": "employee",
                   "data_frame": df,
                   "to_metric": ["id_str"],
                   "to_attribute": ["id_int", "salary", "age"]}]

        DATASETNAME = "Employees"
        DESCRIPTION = "Table of employee data."

        model = Model(tables=tables, name=DATASETNAME, description=DESCRIPTION)

        self.assertEqual(len(model.attributes), 6)
        self.assertEqual(len(model.metrics), 2)

    def test_data_param_no_list(self):
        tables = make_df()

        DATASETNAME = "Employees"
        DESCRIPTION = "Table of employee data."

        self.assertRaises(TypeError, Model, tables=tables, name=DATASETNAME, description=DESCRIPTION)

    def test_data_param_empty_list(self):
        df = make_df()
        tables = [df]

        DATASETNAME = "Employees"
        DESCRIPTION = "Table of employee data."

        self.assertRaises(TypeError, Model, tables=tables, name=DATASETNAME, description=DESCRIPTION)

    def test_data_param_bad_key(self):
        df = make_df()
        tables = [{"wrong1": df,
                   "wrong2": "employees"}]

        DATASETNAME = "Employees"
        DESCRIPTION = "Table of employee data."

        self.assertRaises(ValueError, Model, tables=tables, name=DATASETNAME, description=DESCRIPTION)

    def test_long_dataset_description(self):
        df = make_df()
        tables = [{"table_name": "employee",
                   "data_frame": df}]

        DATASETNAME = "Employees"
        DESCRIPTION = "Lorem ipsum dolor sit amet, consectetur adipiscing elit, " \
                      "sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. " \
                      "Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris " \
                      "nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in " \
                      "reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla " \
                      "pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa " \
                      "qui officia deserunt mollit anim id est laborum."

        self.assertRaises(ValueError, Model, tables=tables, name=DATASETNAME, description=DESCRIPTION)

    def test_folder_id_none(self):
        df = make_df()
        tables = [{"table_name": "employee",
                   "data_frame": df}]

        DATASETNAME = "Employees"
        DESCRIPTION = "Table of employee data."

        model = Model(tables=tables, name=DATASETNAME, description=DESCRIPTION)
        self.assertEqual(model.folder_id, "")

    def test_folder_id_not_none(self):
        df = make_df()
        tables = [{"table_name": "employee",
                   "data_frame": df}]

        DATASETNAME = "Employees"
        DESCRIPTION = "Table of employee data."
        FOLDERID = 'folder123456'

        model = Model(tables=tables, name=DATASETNAME, description=DESCRIPTION, folder_id=FOLDERID)
        self.assertEqual(model.folder_id, FOLDERID)


if __name__ == '__main__':
    unittest.main()



import unittest
from unittest.mock import patch
from mstrio.api import authentication
from mstrio import microstrategy

USERNAME = "user"
PASSWORD = "password"
PROJECT_NAME = "MicroStrategy Tutorial"
BASE_URL = "https://env-xxxxx.customer.cloud.microstrategy.com/MicroStrategyLibrary/api"


class TestAuthentication(unittest.TestCase):

    def setUp(self):
        self.USERNAME = "user"
        self.PASSWORD = "password"
        self.PROJECT_NAME = "Project Name"
        self.BASE_URL = "https://acme.microstrategy.com/MicroStrategyLibrary/api"
        self.COOKIES = "test-cookie"
        self.APPCODE = 64

    @patch('mstrio.api.authentication.requests.post')
    def test_login(self, mock_post):
        conn = microstrategy.Connection(base_url=self.BASE_URL, username=self.USERNAME,
                                        password=self.PASSWORD, project_name=self.PROJECT_NAME)

        mock_post.return_value.status_code = 200

        response = authentication.login(conn)

        self.assertEqual(response.status_code, 200)

    @patch('mstrio.api.authentication.requests.post')
    def test_login_app_code_present(self, mock_post):
        """Checks that application code is present in request body."""
        conn = microstrategy.Connection(base_url=self.BASE_URL, username=self.USERNAME,
                                        password=self.PASSWORD, project_name=self.PROJECT_NAME)

        mock_post.return_value.status_code = 200
        response = authentication.login(conn)

        self.assertEqual(response.status_code, 200)
        self.assertEqual(mock_post.call_args[1]['data']["applicationType"], self.APPCODE)

    def test_login_app_code_static(self, mock_post):
        """Checks that application code is 64."""
        conn = microstrategy.Connection(base_url=self.BASE_URL, username=self.USERNAME,
                                        password=self.PASSWORD, project_name=self.PROJECT_NAME)

        mock_post.return_value.status_code = 200
        response = authentication.login(conn)

        self.assertEqual(response.status_code, 200)
        self.assertEqual(conn.application_code, self.APPCODE)
        self.assertEqual(mock_post.call_args[1]['data']["applicationType"], self.APPCODE)

    @patch('mstrio.api.authentication.requests.post')
    def test_login_app_code_dynamic(self, mock_post):
        """Tests that changing the app code works."""
        conn = microstrategy.Connection(base_url=self.BASE_URL, username=self.USERNAME,
                                        password=self.PASSWORD, project_name=self.PROJECT_NAME)

        app_code = 99
        conn.application_code = app_code

        mock_post.return_value.status_code = 200
        response = authentication.login(conn)

        self.assertEqual(response.status_code, 200)
        self.assertEqual(conn.application_code, app_code)
        self.assertEqual(mock_post.call_args[1]['data']["applicationType"], app_code)

    @patch('mstrio.api.authentication.requests.post')
    def test_logout(self, mock_post):
        """Validate logout method and that cookies are passed in request."""
        conn = microstrategy.Connection(base_url=self.BASE_URL, username=self.USERNAME,
                                        password=self.PASSWORD, project_name=self.PROJECT_NAME)
        conn.cookies = self.COOKIES

        mock_post.return_value.status_code = 200
        response = authentication.logout(conn)

        self.assertEqual(response.status_code, 200)
        self.assertEqual(mock_post.call_args[1]['cookies'], self.COOKIES)

    @patch('mstrio.api.authentication.requests.put')
    def test_sessions(self, mock_put):
        conn = microstrategy.Connection(base_url=self.BASE_URL, username=self.USERNAME,
                                        password=self.PASSWORD, project_name=self.PROJECT_NAME)

        mock_put.return_value.status_code = 200

        response = authentication.sessions(conn)

        self.assertEqual(response.status_code, 200)


if __name__ == '__main__':
    unittest.main()



#TODO (zrogala): Consider changing the way possible errors are handled (deprecate isinstance).
import requests


def report(connection, report_id, verbose=False):
    """
    Get the definition of a specific report, including attributes and metrics. This in-memory report definition provides information about all available objects without actually running any data query/report. The results can be used by other requests to help filter large datasets and retrieve values dynamically, helping with performance and scalability.

    Args:
        connection: MicroStrategy REST API connection object
        report_id (str): Unique ID of the report you wish to extract information from.
        verbose (bool): Verbosity of request response; defaults to False.

    Returns:
        Complete HTTP response object.
    """

    headers = {'X-MSTR-AuthToken': connection.auth_token, 'X-MSTR-ProjectID': connection.project_id}
    response = requests.get(connection.base_url + '/reports/' + report_id,
                            headers=headers,
                            cookies=connection.cookies,
                            verify=connection.ssl_verify)
    if verbose:
        print(response.url)
    return response


def report_instance(connection, report_id, body={}, offset=0, limit=1000, verbose=False):
    """
    Get the results of a newly created report instance. This in-memory report instance can be used by other requests.

    Args:
        connection: MicroStrategy REST API connection object.
        report_id (str): Unique ID of the report you wish to extract information from.
        offset (int, optional): Starting point within the collection of returned results. Default is 0.
        limit (int, optional): Used to control data extract behavior on datasets which have a large number of rows.
            The default is 1000. As an example, if the dataset has 50,000 rows, this function will incrementally
            extract all 50,000 rows in 1,000 row chunks. Depending on system resources, using a higher limit setting
            (e.g. 10,000) may reduce the total time required to extract the entire dataset.
        verbose (bool): Verbosity of request response; defaults to False.

    Returns:
        Complete HTTP response object.
    """

    response = requests.post(url=connection.base_url + '/reports/' + report_id + '/instances/',
                             headers={'X-MSTR-AuthToken': connection.auth_token,
                                      'X-MSTR-ProjectID': connection.project_id},
                             json=body,
                             cookies=connection.cookies,
                             params={'offset': offset,
                                     'limit': limit},
                             verify=connection.ssl_verify)
    if verbose:
        print(response.url)
    return response


def report_instance_id(connection, report_id, instance_id, offset, limit, verbose=False):
    """
    Get the results of a previously created report instance, using the in-memory report instance created by a POST /reports/{reportId}/instances request.

    Args:
        connection: MicroStrategy REST API connection object
        report_id (str): Unique ID of the report you wish to extract information from.
        instance_id (str): Unique ID of the in-memory instance of a published report.
        offset (int): Optional. Starting point within the collection of returned results. Default is 0.
        limit (int): Optional. Used to control data extract behavior on datasets which have a large number of rows.
            The default is 1000. As an example, if the dataset has 50,000 rows, this function will incrementally
            extract all 50,000 rows in 1,000 row chunks. Depending on system resources, using a higher limit setting
            (e.g. 10,000) may reduce the total time required to extract the entire dataset.
        verbose (bool): Verbosity of request response; defaults to False

    Returns:
        Complete HTTP response object.
    """

    response = requests.get(url=connection.base_url + '/reports/' + report_id + '/instances/' + instance_id,
                            headers={'X-MSTR-AuthToken': connection.auth_token,
                                     'X-MSTR-ProjectID': connection.project_id},
                            cookies=connection.cookies,
                            params={'offset': offset,
                                    'limit': limit},
                            verify=connection.ssl_verify)
    if verbose:
        print(response.url)
    return response


def report_single_attribute_elements(connection, report_id, attribute_id, offset=0, limit=25000, verbose=False):
    """
    Get elements of a specific attribute of a specific report.

    Args:
        connection: MicroStrategy REST API connection object.
        report_id (str): Unique ID of the report you wish to extract information from.
        attribute_id (str): Unique ID of the attribute in the report.
        verbose (bool): Verbosity of request response; defaults to False.
    Returns:
        Complete HTTP response object
    """
    response = requests.get(url=connection.base_url + '/reports/' + report_id + '/attributes/' + attribute_id + '/elements',
                            headers={'X-MSTR-AuthToken': connection.auth_token,
                                     'X-MSTR-ProjectID': connection.project_id},
                            cookies=connection.cookies,
                            params={'offset': offset,
                                     'limit': limit},
                            verify=connection.ssl_verify)

    if verbose:
        print(response.url)
    return response



import unittest
from unittest.mock import patch
from mstrio.api import datasets
from mstrio import microstrategy

USERNAME = "user"
PASSWORD = "password"
PROJECT_NAME = "MicroStrategy Tutorial"
BASE_URL = "https://env-xxxxx.customer.cloud.microstrategy.com/MicroStrategyLibrary/api"
JSON_BODY = "testjson"
DATASET_ID = "1501C7F44AEFE6B66FE1BDB1954EF12D"
TABLE_ID = "1501C7F44AEFE6B66FE1BDB1954EF12D"
TABLE_NAME = "MYTABLE"
UPDATE_POLICY = "update"


class TestDatasets(unittest.TestCase):

    @patch('mstrio.api.datasets.requests.post')
    def test_create_dataset(self, mock_post):

        conn = microstrategy.Connection(base_url=BASE_URL, username=USERNAME,
                                        password=PASSWORD, project_name=PROJECT_NAME)

        mock_post.return_value.status_code = 200

        response = datasets.create_dataset(conn, body=JSON_BODY)

        self.assertEqual(response.status_code, 200)

    @patch('mstrio.api.datasets.requests.patch')
    def test_update_dataset(self, mock_patch):

        conn = microstrategy.Connection(base_url=BASE_URL, username=USERNAME,
                                        password=PASSWORD, project_name=PROJECT_NAME)

        mock_patch.return_value.status_code = 200

        response = datasets.update_dataset(conn, dataset_id=DATASET_ID, table_name=TABLE_NAME,
                                           update_policy=UPDATE_POLICY, body=JSON_BODY)

        self.assertEqual(response.status_code, 200)

    @patch('mstrio.api.datasets.requests.delete')
    def test_delete_dataset(self, mock_delete):
        conn = microstrategy.Connection(base_url=BASE_URL, username=USERNAME,
                                        password=PASSWORD, project_name=PROJECT_NAME)

        mock_delete.return_value.status_code = 200

        response = datasets.delete_dataset(conn, dataset_id=DATASET_ID)

        self.assertEqual(response.status_code, 200)

if __name__ == '__main__':
    unittest.main()



import requests


def dataset_definition(connection, dataset_id, fields=None, verbose=False):
    """Get the definition of a dataset.

    Args:
        connection (object): MicroStrategy connection object returned by `microstrategy.Connection()`.
        dataset_id (str): Identifier of a pre-existing dataset. Used when updating a pre-existing dataset.
        fields(list, optional): Specifies object types to be returned. Possible values include tables, columns,
            attributes, and metrics. If no value is set, attributes and metrics are returned.
        verbose (bool, optional): Verbosity of server responses; defaults to False.

    Returns:
        HTTP response object returned by the MicroStrategy REST server

    """
    response = requests.get(url=connection.base_url + '/datasets/' + dataset_id,
                            headers={'X-MSTR-AuthToken': connection.auth_token,
                                     'X-MSTR-ProjectID': connection.project_id},
                            params={'fields': fields},
                            cookies=connection.cookies,
                            verify=connection.ssl_verify)
    if verbose:
        print(response.url)
    return response


def create_dataset(connection, body, verbose=False):
    """Create a single-table dataset from external data uploaded to the MicroStrategy Intelligence Server.

    Args:
        connection (object): MicroStrategy connection object returned by `microstrategy.Connection()`.
        body (str): JSON-formatted definition of the dataset. Generated by `utils.formjson()`.
        verbose (bool, optional): Verbosity of server responses; defaults to False.

    Returns:
        HTTP response object returned by the MicroStrategy REST server

    """
    response = requests.post(url=connection.base_url + '/datasets',
                             headers={'X-MSTR-AuthToken': connection.auth_token,
                                      'X-MSTR-ProjectID': connection.project_id},
                             cookies=connection.cookies,
                             json=body,
                             verify=connection.ssl_verify)

    if verbose:
        print(response.url)
    return response


def update_dataset(connection, dataset_id, table_name, update_policy, body, verbose=False, table_id=None):
    """Update a single-table dataset with external data uploaded to the MicroStrategy Intelligence Server.

    Args:
        connection (object): MicroStrategy connection object returned by `microstrategy.Connection()`.
        dataset_id (str): Identifier of a pre-existing dataset. Used when updating a pre-existing dataset.
        table_id (str): Identifier of the table to update within the MicroStrategy dataset.
        update_policy (str):  Update operation type: 'Add' (inserts new, unique rows), 'Update' (updates data in
            existing rows and columns), 'Upsert' (updates existing data and inserts new rows), 'Replace' (similar to
            truncate, replaces the existing data with new data).
        body (str): JSON-formatted definition of the dataset. Generated by `utils.formjson()`.
        verbose (bool, optional): Verbosity of server responses; defaults to False.

    Returns:
        HTTP response object returned by the MicroStrategy REST server

    """
    response = requests.patch(url=connection.base_url + '/datasets/' + dataset_id + '/tables/' + table_name,
                              headers={'X-MSTR-AuthToken': connection.auth_token,
                                       'X-MSTR-ProjectID': connection.project_id,
                                       'updatePolicy': update_policy},
                              cookies=connection.cookies,
                              json=body,
                              verify=connection.ssl_verify)
    if verbose:
        print(response.url)
    return response


def delete_dataset(connection, dataset_id, verbose=False):
    """Delete a dataset previously created using the REST API.

    Args:
        connection (object): MicroStrategy connection object returned by `microstrategy.Connection()`.
        dataset_id (str): Identifier of a pre-existing dataset. Used when updating a pre-existing dataset.
        verbose (bool, optional): Verbosity of server responses; defaults to False.

    Returns:
        HTTP response object returned by the MicroStrategy REST server

    """
    response = requests.delete(url=connection.base_url + '/objects/' + dataset_id + '?type=3',
                               headers={'X-MSTR-AuthToken': connection.auth_token,
                                        'X-MSTR-ProjectID': connection.project_id},
                               cookies=connection.cookies, verify=connection.ssl_verify)

    if verbose:
        print(response.url)
    return response


def create_multitable_dataset(connection, body, verbose=False):
    """Create the definition of a multi-table dataset.

    Args:
        connection (object): MicroStrategy connection object returned by `microstrategy.Connection()`.
        body (dict): JSON-formatted payload containing the body of the request.
        verbose (bool, optional): Verbosity of server responses; defaults to False.

    Returns:
        HTTP response object returned by the MicroStrategy REST server

    """
    response = requests.post(url=connection.base_url + '/datasets/models',
                             headers={'X-MSTR-AuthToken': connection.auth_token,
                                      'X-MSTR-ProjectID': connection.project_id},
                             cookies=connection.cookies,
                             json=body,
                             verify=connection.ssl_verify)

    if verbose:
        print(response.url)
    return response


def upload_session(connection, dataset_id, body, verbose=False):
    """Create a multi-table dataset upload session.

    Args:
        connection (object): MicroStrategy connection object returned by `microstrategy.Connection()`.
        dataset_id (str): Identifier of a pre-existing dataset. Used when updating a pre-existing dataset.
        body (dict): JSON-formatted payload containing the body of the request.
        verbose (bool, optional): Verbosity of server responses; defaults to False.

    Returns:
        HTTP response object returned by the MicroStrategy REST server

    """

    response = requests.post(url=connection.base_url + '/datasets/' + dataset_id + '/uploadSessions',
                             headers={'X-MSTR-AuthToken': connection.auth_token,
                                      'X-MSTR-ProjectID': connection.project_id},
                             cookies=connection.cookies,
                             json=body,
                             verify=connection.ssl_verify)
    if verbose:
        print(response.url)
    return response


def upload(connection, dataset_id, session_id, body, verbose=False):
    """Upload data to a multi-table dataset.

    Args:
        connection (object): MicroStrategy connection object returned by `microstrategy.Connection()`.
        dataset_id (str): Identifier of a pre-existing dataset. Used when updating a pre-existing dataset.
        session_id (str): Identifer of the server session used for collecting uploaded data.
        body (dict): JSON-formatted payload containing the body of the request.
        verbose (bool, optional): Verbosity of server responses; defaults to False.

    Returns:
        HTTP response object returned by the MicroStrategy REST server

    """

    response = requests.put(url=connection.base_url + '/datasets/' + dataset_id + '/uploadSessions/' + session_id,
                            headers={'X-MSTR-AuthToken': connection.auth_token,
                                     'X-MSTR-ProjectID': connection.project_id},
                            cookies=connection.cookies,
                            json=body,
                            verify=connection.ssl_verify)
    if verbose:
        print(response.url)
    return response


def publish(connection, dataset_id, session_id, verbose=False):
    """Publish a multi-table dataset.

    Args:
        connection (object): MicroStrategy connection object returned by `microstrategy.Connection()`.
        dataset_id (str): Identifier of a pre-existing dataset. Used when updating a pre-existing dataset.
        session_id (str): Identifer for the server session used for collecting uploaded data.
        verbose (bool, optional): Verbosity of server responses; defaults to False.

    Returns:
        HTTP response object returned by the MicroStrategy REST server

    """

    response = requests.post(url=connection.base_url + '/datasets/' + dataset_id + '/uploadSessions/' + session_id + '/publish',
                             headers={'X-MSTR-AuthToken': connection.auth_token,
                                      'X-MSTR-ProjectID': connection.project_id},
                             cookies=connection.cookies,
                             verify=connection.ssl_verify)
    if verbose:
        print(response.url)
    return response


def publish_status(connection, dataset_id, session_id, verbose=False):
    """Get multi-table dataset publication status.

    Args:
        connection (object): MicroStrategy connection object returned by `microstrategy.Connection()`.
        dataset_id (str): Identifier of a pre-existing dataset. Used when updating a pre-existing dataset.
        session_id (str): Identifer for the server session used for collecting uploaded data.
        verbose (bool, optional): Verbosity of server responses; defaults to False.

    Returns:
        HTTP response object returned by the MicroStrategy REST server

    """

    response = requests.get(url=connection.base_url + '/datasets/' + dataset_id + '/uploadSessions/' + session_id + '/publishStatus',
                            headers={'X-MSTR-AuthToken': connection.auth_token,
                                     'X-MSTR-ProjectID': connection.project_id},
                            cookies=connection.cookies,
                            verify=connection.ssl_verify)
    if verbose:
        print(response.url)
    return response


def publish_cancel(connection, dataset_id, session_id, verbose=False):
    """Delete a multi-table dataset upload session and cancel publication.

    Args:
        connection (object): MicroStrategy connection object returned by `microstrategy.Connection()`.
        dataset_id (str): Identifier of a pre-existing dataset. Used when updating a pre-existing dataset.
        session_id (str): Identifer for the server session used for collecting uploaded data.
        verbose (bool, optional): Verbosity of server responses; defaults to False.

    Returns:
        HTTP response object returned by the MicroStrategy REST server

    """

    response = requests.delete(url=connection.base_url + '/datasets/' + dataset_id + '/uploadSessions/' + session_id,
                               headers={'X-MSTR-AuthToken': connection.auth_token,
                                        'X-MSTR-ProjectID': connection.project_id},
                               cookies=connection.cookies,
                               verify=connection.ssl_verify)
    if verbose:
        print(response.url)
    return response



import requests


def projects(connection, verbose=False):
    """
    Args:
        connection: MicroStrategy REST API connection object
        verbose (bool, optional): Verbosity of server responses; defaults to False.
    Returns:
        Complete HTTP response object
    """

    response = requests.get(url=connection.base_url + '/projects',
                            headers={'X-MSTR-AuthToken': connection.auth_token},
                            cookies=connection.cookies, verify=connection.ssl_verify)
    if verbose:
        print(response.url)

    return response



import requests


def server_status(connection, verbose=False):
    """
     Args:
        connection: MicroStrategy REST API connection object
        verbose (bool, optional): Verbosity of server responses; defaults to False.
    Returns:
        Complete HTTP response object
    """

    response = requests.get(url=connection.base_url + '/status')

    if verbose:
        print(response.url)

    return response



import requests


def cube(connection, cube_id, verbose=False):
    """
    Get the definition of a specific cube, including attributes and metrics. The cube can be either an Intelligent Cube
    or a Direct Data Access (DDA)/MDX cube. The in-memory cube definition provides information about all available
    objects without actually running any data query/report. The results can be used by other requests to help filter
    large datasets and retrieve values dynamically, helping with performance and scalability.

    Args:
        connection: MicroStrategy REST API connection object.
        cube_id (str): Unique ID of the cube you wish to extract information from.
        verbose (bool): Verbosity of request response; defaults to False.

    Returns:
        Complete HTTP response object
    """
    response = requests.get(url=connection.base_url + '/cubes/' + cube_id,
                            headers={'X-MSTR-AuthToken': connection.auth_token,
                                     'X-MSTR-ProjectID': connection.project_id},
                            cookies=connection.cookies,
                            verify=connection.ssl_verify)
    if verbose:
        print(response.url)
    return response


def cube_info(connection, cube_id, verbose=False):
    """
    Get information for specific cubes in a specific project. The cubes can be either Intelligent cubes or Direct
    Data Access (DDA)/MDX cubes. This request returns the cube name, ID, size, status, path, last modification date,
    and owner name and ID.

    Args:
        connection: MicroStrategy REST API connection object.
        cube_id (str): Unique ID of the cube you wish to extract information from.
        verbose (bool): Verbosity of request response; defaults to False.

    Returns:
        Complete HTTP response object.
    """
    response = requests.get(url=connection.base_url + '/cubes/?id=' + cube_id,
                            headers={'X-MSTR-AuthToken': connection.auth_token,
                                     'X-MSTR-ProjectID': connection.project_id},
                            cookies=connection.cookies,
                            verify=connection.ssl_verify)
    if verbose:
        print(response.url)
    return response


def cube_single_attribute_elements(connection, cube_id, attribute_id, offset=0, limit=25000, verbose=False):
    """
    Get elements of a specific attribute of a specific cube.

    Args:
        connection: MicroStrategy REST API connection object.
        cube_id (str): Unique ID of the cube you wish to extract information from.
        attribute_id (str): Unique ID of the attribute in the cube.
        verbose (bool): Verbosity of request response; defaults to False.

    Returns:
        Complete HTTP response object.
    """
    response = requests.get(url=connection.base_url + '/cubes/' + cube_id + '/attributes/' + attribute_id + '/elements',
                            headers={'X-MSTR-AuthToken': connection.auth_token,
                                     'X-MSTR-ProjectID': connection.project_id},
                            cookies=connection.cookies,
                            params={'offset': offset,
                                     'limit': limit},
                            verify=connection.ssl_verify)

    if verbose:
        print(response.url)
    return response


def cube_instance(connection, cube_id, body={}, offset=0, limit=1000, verbose=False):
    """
    Create a new instance of a specific cube. This in-memory instance can be used by other requests.

    Args:
        connection: MicroStrategy REST API connection object.
        cube_id (str): Unique ID of the cube you wish to extract information from.
        offset (int, optional): Starting point within the collection of returned results. Default is 0.
        limit (int, optional): Used to control data extract behavior on datasets which have a large number of rows.
            The default is 1000. As an example, if the dataset has 50,000 rows, this function will incrementally
            extract all 50,000 rows in 1,000 row chunks. Depending on system resources, using a higher limit
            setting (e.g. 10,000) may reduce the total time required to extract the entire dataset.
        verbose (bool, optional): Verbosity of request response; defaults to False.

    Returns:
        Complete HTTP response object.
    """
    response = requests.post(url=connection.base_url + '/cubes/' + cube_id + '/instances',
                             headers={'X-MSTR-AuthToken': connection.auth_token,
                                      'X-MSTR-ProjectID': connection.project_id},
                             json=body,
                             cookies=connection.cookies,
                             params={'offset': offset,
                                     'limit': limit},
                             verify=connection.ssl_verify)
    if verbose:
        print(response.url)
    return response


def cube_instance_id(connection, cube_id, instance_id, offset=0, limit=1000, verbose=False):
    """
    Get the results of a previously created instance for a specific cube, using the in-memory instance created by cube_instance().

    Args:
        connection: MicroStrategy REST API connection object.
        cube_id (str): Unique ID of the cube you wish to extract information from.
        instance_id (str): Unique ID of the in-memory instance of a published cube.
        offset (int, optional): Starting point within the collection of returned results. Default is 0.
        limit (int, optional): Used to control data extract behavior on datasets which have a large number of rows.
            The default is 1000. As an example, if the dataset has 50,000 rows, this function will incrementally
            extract all 50,000 rows in 1,000 row chunks. Depending on system resources, using a higher limit
            setting (e.g. 10,000) may reduce the total time required to extract the entire dataset.
        verbose (bool, optional): Verbosity of request response; defaults to False.

    Returns:
        Complete HTTP response object.

    """
    response = requests.get(url=connection.base_url + '/cubes/' + cube_id + '/instances/' + instance_id,
                            headers={'X-MSTR-AuthToken': connection.auth_token,
                                     'X-MSTR-ProjectID': connection.project_id},
                            cookies=connection.cookies,
                            params={'offset': offset,
                                    'limit': limit},
                            verify=connection.ssl_verify)
    if verbose:
        print(response.url)
    return response


def publish(connection, cube_id, verbose=False):
    """
    Publish a specific cube in a specific project.

    Args:
        connection: MicroStrategy REST API connection object.
        cube_id (str): Unique ID of the cube you wish to publish.
        verbose (bool): Verbosity of request response; defaults to False.
    
    Returns:
        Complete HTTP response object.
    """

    response = requests.post(url=connection.base_url + '/cubes/' + cube_id,
                             headers={'X-MSTR-AuthToken': connection.auth_token,
                                      'X-MSTR-ProjectID': connection.project_id},
                             cookies=connection.cookies,
                             verify=connection.ssl_verify)
    if verbose:
        print(response.url)
    return response


def status(connection, cube_id, verbose=False):
    """
    Get the status of a specific cube in a specific project. The status is returned in 
    HEADER X-MSTR-CubeStatus with a value from EnumDSSCubeStates, which is a bit vector.

    Args:
        connection: MicroStrategy REST API connection object.
        cube_id (str): Unique ID of the cube you wish to extract information from.
        verbose (bool): Verbosity of request response; defaults to False.
    
    Returns:
        Complete HTTP response object.
    """

    response = requests.head(url=connection.base_url + '/cubes/' + cube_id,
                             headers={'X-MSTR-AuthToken': connection.auth_token,
                                      'X-MSTR-ProjectID': connection.project_id},
                             cookies=connection.cookies,
                             verify=connection.ssl_verify)
    if verbose:
        print(response.url)
    return response



# TODO: Q-pass in cube or report object?
# TODO: Q-how would you do this if you knew the attr elem ids ahead of time but did not load into F()?
import warnings


class Filter:
    err_msg_invalid = "Invalid object ID: '{}'"
    err_msg_duplicated = "Duplicate object ID: '{}'"

    def __init__(self, attributes, metrics, attr_elements=None):

        self.attributes = {}
        for a in attributes:
            self.attributes[a["id"]] = {"name": a["name"], "selected": False}

        self.metrics = {}
        for m in metrics:
            self.metrics[m["id"]] = {"name": m["name"], "selected": False}

        self.attr_elems = {}
        if attr_elements is not None:
            for att in attr_elements:
                for el in att["elements"]:
                    self.attr_elems[el["id"]] = {"name": att["attribute_name"],
                                                 "attribute_id": att["attribute_id"],
                                                 "selected": False}

        self.attr_selected = None
        self.metr_selected = None
        self.attr_elem_selected = None

    def select(self, object_id):
        if isinstance(object_id, list):
            for i in object_id:
                self.select(object_id=i)
        else:
            if self.__invalid(object_id):
                raise ValueError(self.err_msg_invalid.format(object_id))

            if self.__duplicated(object_id):
                warnings.warn(self.err_msg_duplicated.format(object_id), UserWarning, stacklevel=2)
            else:
                typ = self.__type(object_id)

                if typ == "attribute":
                    if self.attr_selected is None:
                        self.attr_selected = []
                    self.attributes[object_id]["selected"] = True
                    self.attr_selected.append(object_id)

                if typ == "metric":
                    if self.metr_selected is None:
                        self.metr_selected = []
                    self.metrics[object_id]["selected"] = True
                    self.metr_selected.append(object_id)

                if typ == "element":
                    if self.attr_elem_selected is None:
                        self.attr_elem_selected = []
                    self.attr_elems[object_id]["selected"] = True
                    self.attr_elem_selected.append(object_id)

    def clear(self):
        """Removes all previously chosen objects from the filter."""
        for k, v in self.attributes.items():
            self.attributes[k]["selected"] = False

        for k, v in self.metrics.items():
            self.metrics[k]["selected"] = False

        if self.attr_elems is not None:
            for k, v in self.attr_elems.items():
                self.attr_elems[k]["selected"] = False

        self.attr_selected = None
        self.metr_selected = None
        self.attr_elem_selected = None

    def requested_objects(self):
        ro = {}
        if self.attr_selected is not None:
            ro["attributes"] = [{"id": i} for i in self.attr_selected]

        if self.metr_selected is not None:
            ro["metrics"] = [{"id": i} for i in self.metr_selected]

        return ro

    def view_filter(self):

        if not self.attr_elem_selected:
            return None

        else:
            # build {attribute_id:[element_id, element_id_n]} lookup dict
            lkp = {}
            for s in self.attr_elem_selected:
                if self.attr_elems[s]["attribute_id"] in lkp.keys():
                    lkp[self.attr_elems[s]["attribute_id"]].append(s)
                else:
                    lkp[self.attr_elems.get(s)["attribute_id"]] = [s]

            # build view filter
            opers = []
            for k, v in lkp.items():
                att = {"type": "attribute", "id": k}
                elem = {"type": "elements",
                        "elements": [{"id": _} for _ in v]}
                opers.append({"operator": "In",
                              "operands": [att, elem]})

            if len(opers) > 1:
                vf = {"operator": "And", "operands": [op for op in opers]}
            else:
                vf = opers[0]

            return vf

    def filter_body(self):
        fb = {}
        if self.attr_selected is not None or self.metr_selected is not None:
            fb["requestedObjects"] = self.requested_objects()
        if self.attr_elem_selected:
            fb["viewFilter"] = self.view_filter()

        return fb

    def __type(self, object_id):
        """Look up and return object type from available objects."""

        if object_id in self.attributes.keys():
            return "attribute"

        elif object_id in self.metrics.keys():
            return "metric"

        elif object_id in self.attr_elems.keys():
            return "element"

        else:
            return None

    def __invalid(self, object_id):
        """Check if requested object_id is a valid object id."""

        if object_id in self.attributes.keys():
            return False

        elif object_id in self.metrics.keys():
            return False

        elif object_id in self.attr_elems.keys():
            return False

        else:
            return True

    def __duplicated(self, object_id):
        """Check if requested object_id is already selected."""

        if object_id in self.attributes.keys():
            return self.attributes[object_id]["selected"]

        elif object_id in self.metrics.keys():
            return self.metrics[object_id]["selected"]

        elif object_id in self.attr_elems.keys():
            return self.attr_elems[object_id]["selected"]

        else:
            return None



import unittest
import pandas as pd
from mstrio.utils.formjson import formjson


def make_df():
    # this is the base data frame used for testing
    raw_data = {"id_int": [1, 2, 3, 4, 5],
                "id_str": ["1", "2", "3", "4", "5"],
                "first_name": ["Jason", "Molly", "Tina", "Jake", "Amy"],
                "last_name": ["Miller", "Jacobson", "Turner", "Milner", "Cooze"],
                "age": [42, 52, 36, 24, 73],
                "weight": [100.22, 210.2, 175.1, 155.9, 199.9],
                "state": ["VA", "NC", "WY", "CA", "CA"],
                "salary": [50000, 100000, 75000, 85000, 250000]}
    df = pd.DataFrame(raw_data, columns=["id_int", "id_str", "first_name", "last_name",
                                         "age", "weight", "state", "salary"])
    return df


class TestFormjson(unittest.TestCase):

    def test_column_headers(self):
        column_headers, attribute_list, metric_list = formjson(df=make_df(), table_name='TEST')
        testcase = [{'name': 'id_int', 'dataType': 'INTEGER'},
                    {'name': 'id_str', 'dataType': 'STRING'},
                    {'name': 'first_name', 'dataType': 'STRING'},
                    {'name': 'last_name', 'dataType': 'STRING'},
                    {'name': 'age', 'dataType': 'INTEGER'},
                    {'name': 'weight', 'dataType': 'DOUBLE'},
                    {'name': 'state', 'dataType': 'STRING'},
                    {'name': 'salary', 'dataType': 'INTEGER'}]
        self.assertListEqual(column_headers, testcase)

    def test_attribute_list(self):
        column_headers, attribute_list, metric_list = formjson(df=make_df(), table_name='TEST')
        testcase = [{'name': 'id_str', 'attributeForms': [{'category': 'ID', 'expressions': [{'formula': 'TEST.id_str'}]}]},
                    {'name': 'first_name', 'attributeForms': [{'category': 'ID', 'expressions': [{'formula': 'TEST.first_name'}]}]},
                    {'name': 'last_name', 'attributeForms': [{'category': 'ID', 'expressions': [{'formula': 'TEST.last_name'}]}]},
                    {'name': 'state', 'attributeForms': [{'category': 'ID', 'expressions': [{'formula': 'TEST.state'}]}]}]

        self.assertListEqual(attribute_list, testcase)

    def test_metric_list(self):
        column_headers, attribute_list, metric_list = formjson(df=make_df(), table_name='TEST')
        testcase = [{'name': 'id_int', 'dataType': 'number', 'expressions': [{'formula': 'TEST.id_int'}]},
                    {'name': 'age', 'dataType': 'number', 'expressions': [{'formula': 'TEST.age'}]},
                    {'name': 'weight', 'dataType': 'number', 'expressions': [{'formula': 'TEST.weight'}]},
                    {'name': 'salary', 'dataType': 'number', 'expressions': [{'formula': 'TEST.salary'}]}]

        self.assertListEqual(metric_list, testcase)

    def test_formjson_with_attribute_override(self):

        # in this test, the element 'id_int' from the data frame needs to be converted to a attribute in the data model.
        # to accomplish this, simply check that id_int is present in the model json when it is specified as the override

        column_headers, attribute_list, metric_list = formjson(df=make_df(), table_name='TEST', as_attributes=['id_int'])

        # check that id_int is in attribute list
        attributes = [i['name'] for i in attribute_list]
        self.assertIn('id_int', attributes)

    def test_formjson_with_metric_override(self):

        # in this test, the element 'id_str' from the data frame needs to be converted to a metric in the data model.
        # to accomplish this, simply check that id_str is present in the model json when it is specified as the override

        column_headers, attribute_list, metric_list = formjson(df=make_df(), table_name='TEST', as_metrics=['id_str'])

        # check that id_str is in metric list
        metrics = [i['name'] for i in metric_list]
        self.assertIn('id_str', metrics)

    def test_formjson_with_32_bit_metrics(self):

        # create df and change types as needed
        df = make_df()
        df.weight = df.weight.astype('float32')
        df.id_int = df.id_int.astype('int32')
        df.age = df.age.astype('int32')
        df.salary = df.salary.astype('int32')

        # generate json string model from the test dataframe
        column_headers, attribute_list, metric_list = formjson(df=df, table_name='TEST')
        for col in column_headers:
            if col['name'] == 'weight':
                self.assertEqual(col['dataType'], 'DOUBLE')
            if col['name'] in ['id_int', 'age', 'salary']:
                self.assertEqual(col['dataType'], 'INTEGER')

        # Test that created json for the above metrics as in the metrics from created_json and that types didnt change

    def test_formjson_with_64_bit_metrics(self):

        # create df and change types as needed
        df = make_df()
        df.weight = df.weight.astype('float64')
        df.id_int = df.id_int.astype('int64')
        df.age = df.age.astype('int64')
        df.salary = df.salary.astype('int64')

        # generate json string model from the test dataframe
        column_headers, attribute_list, metric_list = formjson(df=df, table_name='TEST')
        for col in column_headers:
            if col['name'] == 'weight':
                self.assertEqual(col['dataType'], 'DOUBLE')
            if col['name'] in ['id_int', 'age', 'salary']:
                self.assertEqual(col['dataType'], 'INTEGER')


if __name__ == '__main__':
    unittest.main()



import json
import pandas as pd
from tqdm.auto import tqdm

from mstrio.api import datasets
from mstrio.utils.model import Model
from mstrio.utils.encoder import Encoder


class Dataset:
    """Create and update data in MicroStrategy datasets.
    Iteratively build a dataset with `Dataset.add_table()`. Then, create the dataset using `Dataset.create()`. When
    updating data in the dataset, add individual tables to the dataset and define how the dataset should be updated
    on the MicroStrategy server, then call `Dataset.update().`
    Attributes:
        name: Name of the dataset.
        description: Description given to the dataset.
        dataset_id: Unique identifier for the dataset. Used to update a pre-existing dataset or generated after creating
            a new dataset.
        __upload_body: Body of the request used to describe the dataset update operation.
        _session_id: Identifies the data upload session.
    """

    __VALID_POLICY = ['add', 'update', 'replace', 'upsert']
    __MAX_DESC_LEN = 250

    def __init__(self, connection, name=None, description=None, dataset_id=None):
        """Interface for creating, updating, and deleting MicroStrategy in-memory datasets.
        When creating a new dataset, provide a dataset name and an optional description. When updating a pre-existing
        dataset, provide the dataset identifier. Tables are added to the dataset in an
        iterative manner using `add_table()`.
        Args:
            connection: MicroStrategy connection object returned by `microstrategy.Connection()`.
            name (str): Name of the dataset.
            description (str, optional): Description of the dataset. Must be less than or equal to 250 characters.
            dataset_id (str, optional): Identifier of a pre-existing dataset. Used when updating a pre-existing dataset.
        """

        if name is not None:
            self.__check_param_str(name, msg="Dataset name should be a string.")
            self.__check_param_len(name,
                                   msg="Dataset name should be <= {} characters.".format(self.__MAX_DESC_LEN),
                                   length=self.__MAX_DESC_LEN)
        self._name = name

        if description is not None:
            self.__check_param_str(description, msg="Dataset description should be a string.")
            self.__check_param_str(description, msg="Dataset description should be a string.")
            self.__check_param_len(description,
                                   msg="Dataset description should be <= {} characters.".format(self.__MAX_DESC_LEN),
                                   length=self.__MAX_DESC_LEN)
        self._desc = description

        if dataset_id is not None:
            self.__check_param_str(dataset_id, "Dataset ID should be a string.")
        self._dataset_id = dataset_id

        self._connection = connection
        self._tables = []
        self._definition = None
        self._session_id = None
        self._folder_id = None
        self.__upload_body = None

    def add_table(self, name, data_frame, update_policy, to_metric=None, to_attribute=None):
        """Add a Pandas DataFrame to a collection of tables which are later used to update the MicroStrategy dataset.
        Args:
            name (str): Logical name of the table that is visible to users of the dataset in MicroStrategy.
            data_frame (:obj:`pandas.DataFrame`): Pandas DataFrame to add or update.
            update_policy (str): Update operation to perform. One of 'add' (inserts new, unique rows), 'update' (updates
                data in existing rows and columns), 'upsert' (updates existing data and inserts new rows), or 'replace'
                (replaces the existing data with new data).
            to_metric (optional, :obj:`list` of str): By default, Python numeric data types are treated as metrics in
                the MicroStrategy dataset while character and date types are treated as attributes. For example, a
                column of integer-like strings ("1", "2", "3") would, by default, be an attribute in the newly created
                dataset. If the intent is to format this data as a metric, provide the respective column name as
                a string in a list to the `to_metric` parameter.
            to_attribute (optional, :obj:`list` of str): Logical opposite of `to_metric`. Helpful for formatting an
                integer-based row identifier as a primary key in the dataset.
        """

        if not isinstance(data_frame, pd.DataFrame):
            raise TypeError("data_frame must be a valid Pandas DataFrame.")

        if update_policy.lower() not in self.__VALID_POLICY:
            raise ValueError("Update policy must be one of '{}', '{}', '{},' or '{}'.".format(*self.__VALID_POLICY))

        table = {"table_name": name,
                 "data_frame": data_frame,
                 "update_policy": update_policy.lower()}

        # check for duplicate column names when overriding attribute and metric names
        if to_attribute and to_metric and any(col in to_attribute for col in to_metric):
            raise ValueError("Column name(s) present in `to_attribute` also present in 'to_metric'.")

        if to_attribute is not None:
            if any([col for col in to_attribute if col not in data_frame.columns]):
                raise ValueError("Column name(s) in `to_attribute` were not found in `DataFrame.columns`.")
            else:
                table["to_attribute"] = to_attribute

        if to_metric is not None:
            if any([col for col in to_metric if col not in data_frame.columns]):
                raise ValueError("Column name(s) in `to_metric` were not found in `DataFrame.columns`.")
            else:
                table["to_metric"] = to_metric

        self._tables.append(table)

    def create(self, folder_id=None, auto_upload=True, chunksize=100000, progress_bar=True, verbose=False):
        """Creates a new dataset.
        Args:
            folder_id (str, optional): ID of the shared folder that the dataset should be created within. If `None`,
                defaults to the user's My Reports folder.
            auto_upload: If True, automatically uploads the data used to create the dataset definition to the dataset.
                If False, simply creates the dataset but does not upload data to it.
            chunksize (int, optional): Number of rows to transmit to the server with each request when uploading.
            progress_bar(bool, optional): If True (default), show the upload progress bar.
            verbose: If True, prints status information about the dataset upload.
        """

        if folder_id is not None:
            self._folder_id = folder_id
        else:
            self._folder_id = ""

        # generate model of the dataset
        self.__build_model()

        # makes request to create the dataset
        response = datasets.create_multitable_dataset(connection=self._connection, body=self.__model)

        if not response.ok:
            self.__response_handler(response=response, msg="Error creating new dataset model.")
        else:
            response_json = response.json()
            self._dataset_id = response_json['id']

            if verbose:
                print("Created dataset '{}' with ID: '{}'.".format(*[self._name, self._dataset_id]))

        # if desired, automatically upload and publish the data to the new dataset
        if auto_upload:
            self.update(chunksize=chunksize, progress_bar=progress_bar)
            self.publish()

            status = 6  # default initial status
            while status != 1:
                pub = datasets.publish_status(connection=self._connection, dataset_id=self._dataset_id,
                                              session_id=self._session_id)
                if not pub.ok:
                    self.__response_handler(response=pub,
                                            msg="Error publishing the dataset.")
                    break
                else:
                    pub = pub.json()
                    status = pub['status']
                    if status == 1:
                        break

    def update(self, chunksize=100000, progress_bar=True):
        """Updates a dataset with new data.
        Args:
            chunksize (int, optional): Number of rows to transmit to the server with each request.
            progress_bar(bool, optional): If True (default), show the upload progress bar.
        """

        # form request body and create a session for data uploads
        self.__form_upload_body()
        response = datasets.upload_session(connection=self._connection,
                                           dataset_id=self._dataset_id, body=self.__upload_body)

        if not response.ok:
            self.__response_handler(response=response, msg="Error creating new data upload session.")
        else:
            response_json = response.json()
            self._session_id = response_json['uploadSessionId']

            # upload each table
            for ix, _table in enumerate(self._tables):

                _df, _name = _table["data_frame"], _table["table_name"]

                # break the data up into chunks using a generator
                chunks = (_df[i:i + chunksize] for i in range(0, _df.shape[0], chunksize))

                total = _df.shape[0]

                # Count the number of iterations
                it_total = int(total/chunksize) + (total % chunksize != 0)

                pbar = tqdm(chunks, total=it_total, disable=(not progress_bar))
                for index, chunk in enumerate(pbar):
                    if progress_bar:
                        pbar.set_description("Uploading {}/{}".format(ix+1, len(self._tables)))

                    # base64 encode the data
                    encoder = Encoder(data_frame=chunk, dataset_type='multi')
                    b64_enc = encoder.encode

                    # form body of the request
                    body = {"tableName": _name,
                            "index": index + 1,
                            "data": b64_enc}

                    # make request to upload the data
                    response = datasets.upload(connection=self._connection,
                                            dataset_id=self._dataset_id,
                                            session_id=self._session_id,
                                            body=body)

                    if not response.ok:
                        # on error, cancel the previously uploaded data
                        self.__response_handler(response=response, msg="Error uploading data.")
                        datasets.publish_cancel(connection=self._connection,
                                                dataset_id=self._dataset_id,
                                                session_id=self._session_id)

                    if progress_bar:
                        pbar.set_postfix(rows=min((index+1)*chunksize, total))
                pbar.close()
            self._tables = []

    def publish(self):
        """Publish the uploaded data to the selected dataset.
        Returns:
            response: Response from the Intelligence Server acknowledging the publication process.
        """

        response = datasets.publish(connection=self._connection,
                                    dataset_id=self._dataset_id, session_id=self._session_id)

        if not response.ok:
            # on error, cancel the previously uploaded data
            self.__response_handler(response=response, msg="Error publishing uploaded data. Cancelling publication.")
            datasets.publish_cancel(connection=self._connection,
                                    dataset_id=self._dataset_id, session_id=self._session_id)

        return response

    def publish_status(self):
        """Check the status of data that was uploaded to a dataset.
        Returns:
            status: The status of the publication process as a dictionary. In the 'status' key, "1" denotes completion.
        """
        response = datasets.publish_status(connection=self._connection,
                                           dataset_id=self._dataset_id, session_id=self._session_id)
        status = response.json()

        return status

    def delete(self, verbose=False):
        """Delete a dataset that was previously created using the REST API.
         Args:
            verbose: If True, prints status information about the dataset upload.
        """
        response = datasets.delete_dataset(connection=self._connection, dataset_id=self._dataset_id)

        if not response.ok:
            self.__response_handler(response=response,
                                    msg="Error deleting dataset with ID: '{}'".format(self._dataset_id))
        else:
            if verbose:
                print("Successfully deleted dataset ID: '{}'.".format(self._dataset_id))

    def upload_status(self, connection, dataset_id, session_id):
        """Check the status of data that was uploaded to a dataset.
        Args:
            connection: MicroStrategy connection object returned by `microstrategy.Connection()`.
            dataset_id (str): Identifier of a pre-existing dataset.
            session_id (str): Identifer of the server session used for collecting uploaded data.
        """
        response = datasets.publish_status(connection=connection, dataset_id=dataset_id, session_id=session_id)

        self.__response_handler(response=response,
                                msg="Publication status for dataset with ID: '{}':".format(dataset_id))

    def __build_model(self):
        """Create json representation of the dataset."""

        # generate model
        model = Model(tables=self._tables, name=self._name, description=self._desc, folder_id=self._folder_id)
        self.__model = model.get_model()

    def __form_upload_body(self):
        """Form request body for creating an upload session for data uploads"""

        # generate body string
        body = {"tables": [{"name": tbl["table_name"],
                            "updatePolicy": tbl["update_policy"],
                            "columnHeaders": list(tbl["data_frame"].columns)} for tbl in self._tables]}
        self.__upload_body = body

    def __load_definition(self):
        """Load definition of an existing dataset."""

        response = datasets.dataset_definition(connection=self._connection, dataset_id=self._dataset_id)

        if not response.ok:
            self.__response_handler(response=response,
                                    msg="Error loading dataset '{}'. Check dataset ID.".format(self._dataset_id))
        else:
            self._definition = response.json()
            self._name = self._definition['name']
            self._dataset_id = self._definition['id']

    @staticmethod
    def __response_handler(response, msg):
        """Generic error message handler for transactions against datasets.
        Args:
            response: Response object returned by HTTP request.
            msg (str): Message to print in addition to any server-generated error message(s).
        """
        res = json.loads(response.content)
        print(msg)
        print("HTTP %i %s" % (response.status_code, response.reason))
        print("I-Server Error %s, %s" % (res['code'], res['message']))

    @staticmethod
    def __check_param_len(param, msg, length):
        if len(param) >= length:
            raise ValueError(msg)
        else:
            return True

    @staticmethod
    def __check_param_str(param, msg):
        if not isinstance(param, str):
            raise TypeError(msg)
        else:
            return True

    @property
    def name(self):
        return self._name

    @property
    def description(self):
        return self._desc

    @property
    def dataset_id(self):
        return self._dataset_id

    @property
    def upload_body(self):
        return self.__upload_body

    @property
    def session_id(self):
        return self._session_id



import unittest
from unittest.mock import patch
from mstrio.api import cubes
from mstrio import microstrategy

USERNAME = "user"
PASSWORD = "password"
PROJECT_NAME = "MicroStrategy Tutorial"
BASE_URL = "https://env-xxxxx.customer.cloud.microstrategy.com/MicroStrategyLibrary/api"
CUBE_ID = "1"
INSTANCE_ID = "2"


class TestCubes(unittest.TestCase):

    @patch('mstrio.api.cubes.requests.get')
    def test_cube(self, mock_get):

        conn = microstrategy.Connection(base_url=BASE_URL, username=USERNAME,
                                        password=PASSWORD, project_name=PROJECT_NAME)

        mock_get.return_value.status_code = 200

        response = cubes.cube(conn, cube_id=CUBE_ID)

        self.assertEqual(response.status_code, 200)

    @patch('mstrio.api.cubes.requests.post')
    def test_cube_instance(self, mock_post):

        conn = microstrategy.Connection(base_url=BASE_URL, username=USERNAME,
                                        password=PASSWORD, project_name=PROJECT_NAME)

        mock_post.return_value.status_code = 200

        response = cubes.cube_instance(conn, cube_id=CUBE_ID)

        self.assertEqual(response.status_code, 200)

    @patch('mstrio.api.cubes.requests.get')
    def test_cube_instance_id(self, mock_get):

        conn = microstrategy.Connection(base_url=BASE_URL, username=USERNAME,
                                        password=PASSWORD, project_name=PROJECT_NAME)

        mock_get.return_value.status_code = 200

        response = cubes.cube_instance_id(conn, cube_id=CUBE_ID, instance_id=INSTANCE_ID)

        self.assertEqual(response.status_code, 200)


if __name__ == '__main__':
    unittest.main()



import json
import warnings
import pandas as pd
from tqdm.auto import tqdm

from mstrio.api import reports
from mstrio.utils.parsejson import parsejson
from mstrio.utils.filter import Filter


class Report:
    """
    Access, filter, publish, and extract data from in-memory reports.

    Create a Report object to load basic information on a report dataset. Specify subset of report
    to be fetched through Report.apply_filters() and Report.clear_filters(). Fetch dataset through
    Report.to_dataframe() method.

    Attributes:
        connection: MicroStrategy connection object returned by `microstrategy.Connection()`.
        report_id: Identifier of a pre-existing report containing the required data.
    """

    def __init__(self, connection, report_id):
        """
        Initialize an instance of a report

        Args:
            connection: MicroStrategy connection object returned by `microstrategy.Connection()`.
            report_id (str): Identifier of a pre-existing report containing the required data.
        """
        init_pbar = tqdm(desc='Initializing an instance of a report. Please wait...',
                         bar_format='{desc}', leave=False, ncols=285)

        self._connection = connection
        self._report_id = report_id

        # load report information
        self.__definition()

        # load attribute elements
        self.__OFFSET = 0
        self.__attr_elements()

        self._filter = Filter(attributes=self._attributes, metrics=self._metrics, attr_elements=self._attr_elements)

        self._dataframe = None
        init_pbar.close()

    def to_dataframe(self, limit=25000, progress_bar=True):
        """Extract contents of a report instance into a Pandas Data Frame. Formerly `microstrategy.Connection.get_report()`.

        Args:
            limit (int, optional): Used to control data extract behavior on datasets with a large number of rows.
                The default is 25000. As an example, if the dataset has 50,000 rows, get_report() will incrementally
                extract all 50,000 rows in 1,000 row chunks. Depending on system resources, a higher limit (e.g. 10,000)
                may reduce the total time required to extract the entire dataset.
            progress_bar(bool, optional): If True (default), show the upload progress bar.

        Returns:
            Pandas Data Frame containing the report contents
        """
        inst_pbar = tqdm(desc='Connecting to MicroStrategy I-Server. Please wait...',
                         bar_format='{desc}', leave=False, ncols=310)

        # Request a new instance, set instance id
        res = reports.report_instance(connection=self._connection,
                                      report_id=self._report_id,
                                      body=self._filter.filter_body(),
                                      offset=self.__OFFSET,
                                      limit=limit)
        inst_pbar.close()
        if not res.ok:
            msg = "Error getting report contents."
            self.__response_handler(response=res, msg=msg)
        else:
            _instance = res.json()
            _instance_id = _instance['instanceId']

            # Gets the pagination totals from the response object
            _pagination = _instance['result']['data']['paging']

            # If there are more rows to fetch, fetch them
            if _pagination['current'] != _pagination['total']:

                # initialize a list to capture slices from each query, and append the first request's result to the list
                table_data = [parsejson(response=_instance)]

                # Count the number of iterations
                it_total = int(_pagination['total']/limit)+(_pagination['total'] % limit != 0)

                # Fetch add'l rows from this object instance from the intelligence server
                with tqdm(total=it_total, disable=(not progress_bar)) as fetch_pbar:
                    if progress_bar:
                        fetch_pbar.update()
                        fetch_pbar.set_description("Downloading")
                        fetch_pbar.set_postfix(rows=limit)
                    for _offset in range(limit, _pagination['total'], limit):
                        if progress_bar:
                            fetch_pbar.update()
                            fetch_pbar.set_description("Downloading")
                            fetch_pbar.set_postfix(rows=min(_offset+limit, _pagination['total']))
                        response = reports.report_instance_id(connection=self._connection,
                                                              report_id=self._report_id,
                                                              instance_id=_instance_id,
                                                              offset=_offset,
                                                              limit=limit)
                        table_data.append(parsejson(response=response.json()))

                # concatenate and return the list of result data as a data frame
                self._dataframe = pd.concat(table_data).reset_index(drop=True)
            else:
                # otherwise parse the first result and return it as a dataframe
                self._dataframe = parsejson(response=_instance)

            return self._dataframe

    def apply_filters(self, attributes=None, metrics=None, attr_elements=None):
        """Apply filters on the report's objects, so only selected attributes, metrics and attributes' elements
        will be retrieved from Intelligence Server.

        Args:
            attributes (list, optional): attributes to be included in the filter.
            metrics (list, optional): metrics to be included in the filter.
            attr_elements (list, optional): attributes' elements to be included in the filter.
        """
        if attributes:
            self._filter.select(object_id=attributes)

        if attributes == []:
            self._filter.attr_selected = []

        if metrics:
            self._filter.select(object_id=metrics)

        if metrics == []:
            self._filter.metr_selected = []

        if attr_elements is not None:
            self._filter.select(object_id=attr_elements)

    def clear_filters(self):
        """Clear previously set filters, allowing all attributes, metrics, and attribute elements to be retrieved."""
        self._filter.clear()

    def __definition(self):
        """Get the definition of a report, including attributes and metrics. Implements GET /reports/<report_id>"""

        res = reports.report(connection=self._connection, report_id=self._report_id)

        if not res.ok:
            self.__response_handler(response=res, msg="Error getting report definition. Check report ID.")
            return None
        else:
            _definition = res.json()
            self._name = _definition["name"]
            full_attributes = _definition["result"]["definition"]["availableObjects"]["attributes"]
            full_metrics = _definition["result"]["definition"]["availableObjects"]["metrics"]
            self._attributes = [{'name': attr['name'], 'id': attr['id']} for attr in full_attributes]
            self._metrics = [{'name': metr['name'], 'id': metr['id']} for metr in full_metrics]

    def __attr_elements(self, limit=25000):
        """Get elements of report attributes. Implements GET /reports/<report_id>/attributes/<attribute_id>/elements"""

        attr_elements = []
        if self._attributes is not None:

            for attr in self._attributes:
                # Fetch first chunk of attribute elements.
                res = reports.report_single_attribute_elements(connection=self._connection,
                                                               report_id=self._report_id,
                                                               attribute_id=attr['id'],
                                                               offset=self.__OFFSET,
                                                               limit=limit)
                if not res.ok:
                    msg = "Error retrieving attribute '" + attr['name'] + "' elements."
                    self.__response_handler(response=res, msg=msg)

                else:
                    # Get total number of rows from headers.
                    total = int(res.headers['x-mstr-total-count'])
                    # Get attribute elements from the response.
                    elements = res.json()

                    # If total number of elements is bigger than the chunk size (limit), fetch them incrementally.
                    for _offset in range(limit, total, limit):
                        res = reports.report_single_attribute_elements(connection=self._connection,
                                                                       report_id=self._report_id,
                                                                       attribute_id=attr['id'],
                                                                       offset=_offset,
                                                                       limit=limit)
                        elements.extend(res.json())

                    # Append attribute data to the list of attributes.
                    attr_elements.append({"attribute_name": attr['name'],
                                          "attribute_id": attr['id'],
                                          "elements": elements})

            self._attr_elements = attr_elements

    @staticmethod
    def __response_handler(response, msg):
        """Generic error message handler for transactions against datasets.

        Args:
            response: Response object returned by HTTP request.
            msg (str): Message to print in addition to any server-generated error message(s).

        """
        res = json.loads(response.content)
        print(msg)
        print("HTTP %i %s" % (response.status_code, response.reason))
        print("I-Server Error %s, %s" % (res['code'], res['message']))

    @property
    def name(self):
        return self._name

    @property
    def attributes(self):
        return self._attributes

    @property
    def metrics(self):
        return self._metrics

    @property
    def attr_elements(self):
        if not self._attr_elements:
            self.__attr_elements()
        return self._attr_elements

    @property
    def selected_attributes(self):
        return self._filter.attr_selected

    @property
    def selected_metrics(self):
        return self._filter.metr_selected

    @property
    def selected_attr_elements(self):
        return self._filter.attr_elem_selected

    @property
    def dataframe(self):
        if self._dataframe is None:
            warnings.warn("Dataframe not loaded. Retrieve with Report.to_dataframe().", Warning, stacklevel=2)
        return self._dataframe



import unittest
from unittest.mock import patch
from mstrio.api import reports
from mstrio import microstrategy

USERNAME = "user"
PASSWORD = "password"
PROJECT_NAME = "MicroStrategy Tutorial"
BASE_URL = "https://env-xxxxx.customer.cloud.microstrategy.com/MicroStrategyLibrary/api"
REPORT_ID = "1"
INSTANCE_ID = "2"
OFFSET = 0
LIMIT = 1000


class TestReports(unittest.TestCase):

    @patch('mstrio.api.reports.requests.get')
    def test_report(self, mock_get):

        conn = microstrategy.Connection(base_url=BASE_URL, username=USERNAME,
                                        password=PASSWORD, project_name=PROJECT_NAME)

        mock_get.return_value.status_code = 200

        response = reports.report(conn, report_id=REPORT_ID)

        self.assertEqual(response.status_code, 200)

    @patch('mstrio.api.reports.requests.post')
    def test_report_instance(self, mock_post):

        conn = microstrategy.Connection(base_url=BASE_URL, username=USERNAME,
                                        password=PASSWORD, project_name=PROJECT_NAME)

        mock_post.return_value.status_code = 200

        response = reports.report_instance(conn, report_id=REPORT_ID, offset=OFFSET, limit=LIMIT)

        self.assertEqual(response.status_code, 200)

    @patch('mstrio.api.reports.requests.get')
    def test_report_instance_id(self, mock_get):

        conn = microstrategy.Connection(base_url=BASE_URL, username=USERNAME,
                                        password=PASSWORD, project_name=PROJECT_NAME)

        mock_get.return_value.status_code = 200

        response = reports.report_instance_id(conn, report_id=REPORT_ID, instance_id=INSTANCE_ID,
                                              offset=OFFSET, limit=LIMIT)

        self.assertEqual(response.status_code, 200)

if __name__ == '__main__':
    unittest.main()



import unittest
from unittest.mock import Mock, patch
import pandas as pd
from mstrio.dataset import Dataset


def make_df():
    # this is the base data frame used for testing
    raw_data = {"id_int": [1, 2, 3, 4, 5],
                "id_str": ["1", "2", "3", "4", "5"],
                "first_name": ["Jason", "Molly", "Tina", "Jake", "Amy"],
                "last_name": ["Miller", "Jacobson", "Turner", "Milner", "Cooze"],
                "age": [42, 52, 36, 24, 73],
                "weight": [100.22, 210.2, 175.1, 155.9, 199.9],
                "state": ["VA", "NC", "WY", "CA", "CA"],
                "salary": [50000, 100000, 75000, 85000, 250000]}
    df = pd.DataFrame(raw_data, columns=["id_int", "id_str", "first_name", "last_name",
                                         "age", "weight", "state", "salary"])
    return df


class TestDatasets(unittest.TestCase):

    def test_init_null_values(self):
        # Test that null param values are assigned properly when initiated

        ds = Dataset(connection={}, name="__test_name")

        self.assertIsNone(ds.description)
        self.assertIsNone(ds.dataset_id)
        self.assertIsNone(ds._definition)
        self.assertIsNone(ds._session_id)
        self.assertIsNone(ds._folder_id)
        self.assertIsNone(ds.upload_body)
        self.assertEqual(len(ds._tables), 0)

    @patch('mstrio.api.datasets.dataset_definition')
    def test_init_non_null(self, mock_definition):
        # Test that non-null param values are assigned properly when initiated

        __test_name = "TEST"
        __test_desc = "TEST DESCRIPTION"
        __test_ds_id = "id1234567890"
        __definition = {'name': 'test_name', 'id': __test_ds_id}
        mock_definition.return_value = Mock(ok=True)
        mock_definition.return_value.json.return_value = __definition

        ds = Dataset(connection={}, name=__test_name)
        self.assertEqual(ds._name, __test_name)

        ds = Dataset(connection={}, name=__test_name, description=__test_desc)
        self.assertEqual(ds._desc, __test_desc)

        ds = Dataset(connection={}, dataset_id=__test_ds_id)
        self.assertTrue(mock_definition.called)
        self.assertEqual(ds._dataset_id, __test_ds_id)

    def test_add_table(self):
        # Test that adding a table to the dataset increases length of tables property by one

        ds = Dataset(connection={}, name="test_name")

        ds.add_table(name="TEST1", data_frame=make_df(), update_policy="add")
        self.assertEqual(len(ds._tables), 1)

        ds.add_table(name="TEST2", data_frame=make_df(), update_policy="add")
        self.assertEqual(len(ds._tables), 2)

    def test_invalid_update_policy(self):
        # Test that invalid update policy values produces an error

        ds = Dataset(connection={}, name="test_name")
        self.assertRaises(ValueError, ds.add_table, name="TEST", data_frame=make_df(), update_policy="invalid")

    def test_invalid_attr_override(self):
        # Test that attribute override columns names which dont match the source table produces an error

        ds = Dataset(connection={}, name="test_name")
        self.assertRaises(ValueError, ds.add_table, name="TEST", data_frame=make_df(), update_policy="add",
                          to_attribute=["invalid"])

    def test_invalid_metr_override(self):
        # Test that metric override columns names which dont match the source table produces an error

        ds = Dataset(connection={}, name="test_name")
        self.assertRaises(ValueError, ds.add_table, name="TEST", data_frame=make_df(), update_policy="add",
                          to_metric=["invalid"])

    def test_duplicate_col_attr_metr_override(self):
        # Test that duplicate attribute and metric column names when override produces an error

        ds = Dataset(connection={}, name="test_name")
        self.assertRaises(ValueError, ds.add_table, name="TEST", data_frame=make_df(), update_policy="add",
                          to_attribute=["age"], to_metric=["age"])


if __name__ == '__main__':
    unittest.main()



from setuptools import setup, find_packages, dist
import os
import pkg_resources
from setuptools.extern.packaging import version as packaging_version

dist_version = '11.1.4.002' #define the default version
if 'CI_VERSION' in os.environ.keys():
    dist_version = os.environ['CI_VERSION'] # read the CI_VERSION from environment variable

# Patch Version class to preserve original version string
dist.pkg_resources.safe_version = lambda v:v

class NoNormalizeVersion(packaging_version.Version):
    def __init__(self, version):
        self._orig_version = version
        super().__init__(version)

    def __str__(self):
        return self._orig_version

packaging_version.Version = NoNormalizeVersion

setup(name='mstrio-py',
      version=dist_version,
      description='Python interface for the MicroStrategy REST API',
      license='Apache License 2.0',
      url='https://github.com/MicroStrategy/mstrio-py',
      author='Scott Rigney, Peter Ott, Sergio Sainz Palacios, Michal Ciesielski, Zofia Rogala, Ignacy Hologa, Piotr Czyz',
      author_email='srigney@microstrategy.com, ssainz@microstrategy.com, mciesielski@microstrategy.com,'
                    'zrogala@microstrategy.com, ihologa@microstrategy.com, pczyz@microstrategy.com',
      install_requires=[
          'requests',
          'pandas',
          'tqdm',
          'jupyter_contrib_nbextensions',
          'ipywidgets',
          'packaging'
      ],
      packages=find_packages(),
      include_package_data=True)



import unittest
from mstrio.utils.filter import Filter


class TestFilter(unittest.TestCase):

    def setUp(self):
        """Create objects to be used in all test methods."""
        self.attributes = [
            {"name": "row", "id": "37B4899E11E996150AC00080EFC58FB4"},
            {"name": "state", "id": "356F7E6411E996150AC00080EFC58FB4"},
            {"name": "category", "id": "356F88B411E996150AC00080EFC58FB4"}]

        self.metrics = [
            {"name": "sales", "id": "356F904811E996150AC00080EFC58FB4"},
            {"name": "profit", "id": "34A8389A11E99615B08E0080EFC50E5E"}]

        self.elements = [
            {"attribute_name": "row",
             "attribute_id": "37B4899E11E996150AC00080EFC58FB4",
             "elements": [
                 {"id": "37B4899E11E996150AC00080EFC58FB4:1", "formValues": ["1"]},
                 {"id": "37B4899E11E996150AC00080EFC58FB4:2", "formValues": ["2"]},
                 {"id": "37B4899E11E996150AC00080EFC58FB4:3", "formValues": ["3"]}]},
            {"attribute_name": "state",
             "attribute_id": "356F7E6411E996150AC00080EFC58FB4",
             "elements": [
                 {"id": "356F7E6411E996150AC00080EFC58FB4:new york", "formValues": ["new york"]},
                 {"id": "356F7E6411E996150AC00080EFC58FB4:virginia", "formValues": ["virginia"]}]},
            {"attribute_name": "category",
             "attribute_id": "356F88B411E996150AC00080EFC58FB4",
             "elements": [
                 {"id": "356F88B411E996150AC00080EFC58FB4:books", "formValues": ["books"]},
                 {"id": "356F88B411E996150AC00080EFC58FB4:music", "formValues": ["music"]}]}]

        self.attribute_sel = self.attributes[0]["id"]
        self.attribute_sel_list = [i["id"] for i in self.attributes]

        self.metric_sel = self.metrics[0]["id"]
        self.metric_sel_list = [i["id"] for i in self.metrics]

        self.element_sel = self.elements[0]["elements"][0]["id"]
        self.element_sel_list = [i["id"] for e in self.elements for i in e["elements"]]         # all from each attr
        self.element_sel_same_list = [self.elements[0]["elements"][n]["id"] for n in range(2)]  # two from 1st attr

        self.invalid_id = "INVALID"

    def test_init_filter_structure(self):
        """Test that init populates the correct instance properties."""
        f = Filter(attributes=self.attributes, metrics=self.metrics, attr_elements=self.elements)

        # assert types
        self.assertIsInstance(f.attributes, dict)
        self.assertIsInstance(f.metrics, dict)
        self.assertIsInstance(f.attr_elems, dict)

        # these should be empty
        self.assertIsNone(f.attr_selected)
        self.assertIsNone(f.metr_selected)
        self.assertIsNone(f.attr_elem_selected)

    def test_init_object_ids(self):
        """Test that each object id is loaded in filter object properties."""
        f = Filter(attributes=self.attributes, metrics=self.metrics, attr_elements=self.elements)

        for a in self.attributes:
            self.assertIn(a["id"], f.attributes)

        for m in self.metrics:
            self.assertIn(m["id"], f.metrics)

        for ae in self.elements:
            for elem in ae["elements"]:
                self.assertIn(elem["id"], f.attr_elems)

    def test_init_attribute_elements(self):
        """Test that when not loading attribute elements, filter's attribute elements are None."""
        f = Filter(attributes=self.attributes, metrics=self.metrics)
        self.assertIsNotNone(f.attr_elems)
        self.assertIsNone(f.attr_elem_selected)

        f = Filter(attributes=self.attributes, metrics=self.metrics, attr_elements=self.elements)
        self.assertIsNotNone(f.attr_elems)
        self.assertIsNone(f.attr_elem_selected)

    def test_select_attribute_id(self):
        """Test adding an object adds the id to filter property."""

        f = Filter(attributes=self.attributes, metrics=self.metrics, attr_elements=self.elements)
        f.select(object_id=self.attribute_sel)

        self.assertEqual(f.attributes[self.attribute_sel]["selected"], True)
        self.assertIn(self.attribute_sel, f.attr_selected)
        self.assertIsNone(f.metr_selected)
        self.assertIsNone(f.attr_elem_selected)

    def test_select_attribute_id_list(self):
        """Test adding an object via list of ids adds the ids to filter property."""

        f = Filter(attributes=self.attributes, metrics=self.metrics, attr_elements=self.elements)
        f.select(object_id=self.attribute_sel_list)

        for obj_id in self.attribute_sel_list:
            self.assertEqual(f.attributes[obj_id]["selected"], True)
            self.assertIn(obj_id, f.attr_selected)
            self.assertIsNone(f.metr_selected)
            self.assertIsNone(f.attr_elem_selected)

    def test_select_duplicate_attribute_id(self):
        """Tests adding a duplicate id does not add the second id to the selected filter."""

        f = Filter(attributes=self.attributes, metrics=self.metrics, attr_elements=self.elements)
        f.select(object_id=self.attribute_sel)

        # add a duplicate
        with self.assertWarns(Warning):
            f.select(object_id=self.attribute_sel)

        # object id be here
        self.assertIn(self.attribute_sel, f.attr_selected)
        self.assertEqual(len(f.attr_selected), 1)

        # object id shouldnt be here
        self.assertIsNone(f.metr_selected)
        self.assertIsNone(f.attr_elem_selected)

    def test_select_metric_id(self):
        """Test adding an object via list of ids adds the ids to filter property."""

        f = Filter(attributes=self.attributes, metrics=self.metrics, attr_elements=self.elements)
        f.select(object_id=self.metric_sel)

        self.assertEqual(f.metrics[self.metric_sel]["selected"], True)
        self.assertIn(self.metric_sel, f.metr_selected)
        self.assertIsNone(f.attr_selected)
        self.assertIsNone(f.attr_elem_selected)

    def test_select_metric_id_list(self):
        """Test adding an object via list of ids adds the ids to filter property."""

        f = Filter(attributes=self.attributes, metrics=self.metrics, attr_elements=self.elements)
        f.select(object_id=self.metric_sel_list)

        for obj_id in self.metric_sel_list:
            self.assertEqual(f.metrics[obj_id]["selected"], True)
            self.assertIn(obj_id, f.metr_selected)
            self.assertIsNone(f.attr_selected)
            self.assertIsNone(f.attr_elem_selected)

    def test_select_duplicate_metric_id(self):
        """Tests adding a duplicate id does not add the second id to the selected filter."""

        obj_id = self.metrics[0]["id"]
        f = Filter(attributes=self.attributes, metrics=self.metrics, attr_elements=self.elements)
        f.select(object_id=self.metric_sel)

        # add a duplicate; should throw a warning
        with self.assertWarns(Warning):
            f.select(object_id=obj_id)

        # object id be here
        self.assertIn(self.metric_sel, f.metr_selected)
        self.assertEqual(len(f.metr_selected), 1)

        # object id shouldnt be here
        self.assertIsNone(f.attr_selected)
        self.assertIsNone(f.attr_elem_selected)

    def test_select_attr_elem_id(self):
        """Test adding an object via id or list of ids adds the ids to filter property."""

        f = Filter(attributes=self.attributes, metrics=self.metrics, attr_elements=self.elements)
        f.select(object_id=self.element_sel)

        self.assertEqual(f.attr_elems[self.element_sel]["selected"], True)
        self.assertIn(self.element_sel, f.attr_elem_selected)
        self.assertIsNone(f.metr_selected)
        self.assertIsNone(f.attr_selected)

    def test_select_attr_elem_id_list(self):
        """Test adding an object via list of ids adds the ids to filter property."""

        f = Filter(attributes=self.attributes, metrics=self.metrics, attr_elements=self.elements)
        f.select(object_id=self.element_sel_list)

        for obj_id in self.element_sel_list:
            self.assertEqual(f.attr_elems[obj_id]["selected"], True)
            self.assertIn(obj_id, f.attr_elem_selected)
            self.assertIsNone(f.metr_selected)
            self.assertIsNone(f.attr_selected)

    def test_select_duplicate_attr_elem_id(self):
        """Tests adding a duplicate id does not add the second id to the selected filter."""

        f = Filter(attributes=self.attributes, metrics=self.metrics, attr_elements=self.elements)
        f.select(object_id=self.element_sel)

        # add a duplicate
        with self.assertWarns(Warning):
            f.select(object_id=self.element_sel)

        # object id be here
        self.assertIn(self.element_sel, f.attr_elem_selected)
        self.assertEqual(len(f.attr_elem_selected), 1)

        # object id shouldn't be here
        self.assertIsNone(f.attr_selected)
        self.assertIsNone(f.metr_selected)

    def test_select_invalid_object_id(self):
        """Tests adding an invalid object id does not add the object id to selected filters."""

        f = Filter(attributes=self.attributes, metrics=self.metrics, attr_elements=self.elements)

        with self.assertRaises(ValueError):
            f.select(object_id=self.invalid_id)

        self.assertIsNone(f.attr_selected)
        self.assertIsNone(f.metr_selected)
        self.assertIsNone(f.attr_elem_selected)

    def test_clear(self):
        """Test that clearing filters works."""

        obj_id = [self.attribute_sel_list,
                  self.metric_sel_list,
                  self.element_sel_list]

        f = Filter(attributes=self.attributes, metrics=self.metrics, attr_elements=self.elements)
        f.select(obj_id)

        self.assertIsNotNone(f.attr_selected)
        self.assertIsNotNone(f.metr_selected)
        self.assertIsNotNone(f.attr_elem_selected)

        # reset
        f.clear()

        self.assertIsNone(f.attr_selected)
        self.assertIsNone(f.metr_selected)
        self.assertIsNone(f.attr_elem_selected)

    def test_requested_objects_empty_lists(self):
        """Test that empty lists are properly parsed into requestedObjects."""
        f = Filter(attributes=self.attributes, metrics=self.metrics, attr_elements=self.elements)
        f.attr_selected = []
        f.metr_selected = []
        ro = f.requested_objects()

        self.assertIn("attributes", ro)
        self.assertIn("metrics", ro)
        self.assertEqual(ro["attributes"], [])
        self.assertEqual(ro["metrics"], [])

    def test_requested_objects_one_attribute(self):
        """Test that choosing 1 att should return matching requested object in body."""

        f = Filter(attributes=self.attributes, metrics=self.metrics, attr_elements=self.elements)
        f.select(self.attribute_sel)

        ro = f.requested_objects()

        self.assertIn("attributes", ro)
        self.assertNotIn("metrics", ro)
        for i in ro["attributes"]:
            self.assertEqual(i["id"], self.attribute_sel)

    def test_requested_objects_two_attribute(self):
        """Test that choosing 2 att should return matching requested object in body."""

        f = Filter(attributes=self.attributes, metrics=self.metrics, attr_elements=self.elements)
        f.select(self.attribute_sel_list)

        ro = f.requested_objects()

        self.assertIn("attributes", ro)
        self.assertNotIn("metrics", ro)
        for i, obj_id in zip(ro["attributes"], self.attribute_sel_list):
            self.assertEqual(i["id"], obj_id)

    def test_requested_objects_one_metric(self):
        """Test that choosing 1 met should return matching requested object in body."""

        f = Filter(attributes=self.attributes, metrics=self.metrics, attr_elements=self.elements)
        f.select(self.metric_sel)

        ro = f.requested_objects()

        self.assertNotIn("attributes", ro)
        self.assertIn("metrics", ro)
        for i in ro["metrics"]:
            self.assertEqual(i["id"], self.metric_sel)

    def test_requested_objects_two_metric(self):
        """Test that choosing 2 met should return matching requested object in body."""

        f = Filter(attributes=self.attributes, metrics=self.metrics, attr_elements=self.elements)
        f.select(self.metric_sel_list)

        ro = f.requested_objects()

        self.assertNotIn("attributes", ro)
        self.assertIn("metrics", ro)
        for i, obj_id in zip(ro["metrics"], self.metric_sel_list):
            self.assertEqual(i["id"], obj_id)

    def test_requested_objects_both_list(self):
        """Test that adding lists of attributes and metrics yields requested objects with correcty elements. """

        f = Filter(attributes=self.attributes, metrics=self.metrics, attr_elements=self.elements)
        f.select(self.attribute_sel_list)
        f.select(self.metric_sel_list)
        ro = f.requested_objects()

        self.assertIn("attributes", ro)
        for i, obj_id in zip(ro["attributes"], self.attribute_sel_list):
            self.assertEqual(i["id"], obj_id)

        self.assertIn("metrics", ro)
        for i, obj_id in zip(ro["metrics"], self.metric_sel_list):
            self.assertEqual(i["id"], obj_id)

    def test_view_filter_none(self):
        """Test that adding no attributes elements yields None view filter."""

        f = Filter(attributes=self.attributes, metrics=self.metrics, attr_elements=self.elements)

        # should be none if none are obj_id
        vf = f.view_filter()
        self.assertIsNone(vf)

    def test_view_filter_not_none(self):
        """Test that adding no attributes elements yields None view filter."""

        f = Filter(attributes=self.attributes, metrics=self.metrics, attr_elements=self.elements)

        # should not be none if some are obj_id
        f.select(self.element_sel)

        vf = f.view_filter()
        self.assertIsNotNone(vf)

    def test_view_filter_same_attrib_one_element(self):
        """Test that choosing 1 att elem should return matching view filter in body."""

        f = Filter(attributes=self.attributes, metrics=self.metrics, attr_elements=self.elements)
        f.select(self.element_sel)
        vf = f.view_filter()

        self.assertIn("operator", vf)
        self.assertEqual(vf["operator"], "In")
        self.assertIn("operands", vf)

        att, elems = vf["operands"]

        self.assertEqual(att["type"], "attribute")
        self.assertEqual(att["id"], f.attr_elems.get(self.element_sel)["attribute_id"])
        self.assertEqual(elems["type"], "elements")

        for elem, obj_id in zip(elems["elements"], self.element_sel):
            self.assertEqual(elem["id"], self.element_sel)

    def test_view_filter_same_attrib_two_element(self):
        """Test that choosing 2 att elems from same attribute should return matching view filter in body."""

        f = Filter(attributes=self.attributes, metrics=self.metrics, attr_elements=self.elements)
        f.select(self.element_sel_same_list)
        vf = f.view_filter()

        self.assertIn("operator", vf)
        self.assertEqual(vf["operator"], "In")  # should be In because elems are for same parent attr
        self.assertIn("operands", vf)

        att, elems = vf["operands"]

        self.assertEqual(att["type"], "attribute")
        self.assertEqual(elems["type"], "elements")

        for obj_id in self.element_sel_same_list:
            self.assertEqual(att["id"], f.attr_elems.get(obj_id)["attribute_id"])

        for elem, obj_id in zip(elems["elements"], self.element_sel_same_list):
            self.assertEqual(elem["id"], obj_id)

    def test_view_filter_multi_attrib_element(self):
        """Test that choosing att elem across attributes forms a correct view filter body."""

        f = Filter(attributes=self.attributes, metrics=self.metrics, attr_elements=self.elements)

        f.select(self.element_sel_list)
        vf = f.view_filter()

        self.assertEqual(vf["operator"], "And")  # should be And because elements span parent attributes
        self.assertIn("operator", vf)
        self.assertIn("operands", vf)
        self.assertIsInstance(vf["operands"], list)

        for operand in vf["operands"]:

            self.assertIn("operator", operand)
            self.assertIn("operands", operand)
            self.assertEqual(operand["operator"], "In")  # should be In because elems are for same parent attr

            att, elems = operand["operands"]

            self.assertEqual(att["type"], "attribute")
            self.assertEqual(elems["type"], "elements")

            for obj_id in elems["elements"]:
                self.assertEqual(att["id"], f.attr_elems.get(obj_id["id"])["attribute_id"])

            for elem, obj_id in zip(elems["elements"], elems["elements"]):
                self.assertEqual(elem["id"], obj_id["id"])

    def test_filter_body_keys(self):
        """Test correctness of filter body."""

        f = Filter(attributes=self.attributes, metrics=self.metrics, attr_elements=self.elements)

        # it should be none
        self.assertIsNotNone(f.filter_body())

        # it should have requested objects
        f.select(self.attribute_sel)
        self.assertIn("requestedObjects", f.filter_body())
        self.assertNotIn("viewFilter", f.filter_body())
        f.clear()

        f.select(self.metric_sel)
        self.assertIn("requestedObjects", f.filter_body())
        self.assertNotIn("viewFilter", f.filter_body())
        f.clear()

        # it should have view filter
        f.select(self.element_sel)
        self.assertIn("viewFilter", f.filter_body())
        self.assertNotIn("requestedObjects", f.filter_body())
        f.clear()

        # it should have requested objects and view filter
        f.select(self.attribute_sel)
        f.select(self.metric_sel)
        f.select(self.element_sel)
        self.assertIn("requestedObjects", f.filter_body())
        self.assertIn("viewFilter", f.filter_body())

    def test_filter_body_attribute_list(self):
        """Test for presence of attribute ids in the filter body."""

        f = Filter(attributes=self.attributes, metrics=self.metrics, attr_elements=self.elements)
        f.select(self.attribute_sel_list)
        ro = f.requested_objects()
        fb = f.filter_body()

        self.assertListEqual(fb["requestedObjects"]["attributes"], ro["attributes"])

    def test_filter_body_metric_list(self):
        """Test for presence of metric ids in the filter body."""

        f = Filter(attributes=self.attributes, metrics=self.metrics, attr_elements=self.elements)
        f.select(self.metric_sel_list)
        ro = f.requested_objects()
        fb = f.filter_body()

        self.assertListEqual(fb["requestedObjects"]["metrics"], ro["metrics"])

    def test_filter_body_attribute_element_list(self):
        """Test for presence of attribute element ids in the filter body."""

        f = Filter(attributes=self.attributes, metrics=self.metrics, attr_elements=self.elements)
        f.select(self.element_sel_list)
        vf = f.view_filter()
        fb = f.filter_body()

        self.assertDictEqual(fb["viewFilter"], vf)


if __name__ == '__main__':
    unittest.main()



import unittest
import pandas as pd
from mstrio.utils.encoder import Encoder
from base64 import b64decode
import json


def make_df():
    # this is the base data frame used for testing
    raw_data = {"id_int": [1, 2, 3, 4, 5],
                "id_str": ["1", "2", "3", "4", "5"],
                "first_name": ["Jason", "Molly", "Tina", "Jake", "Amy"],
                "last_name": ["Miller", "Jacobson", "Turner", "Milner", "Cooze"],
                "age": [42, 52, 36, 24, 73],
                "weight": [100.22, 210.2, 175.1, 155.9, 199.9],
                "state": ["VA", "NC", "WY", "CA", "CA"],
                "salary": [50000, 100000, 75000, 85000, 250000]}
    df = pd.DataFrame(raw_data, columns=["id_int", "id_str", "first_name", "last_name",
                                         "age", "weight", "state", "salary"])
    return df


class TestEncoder(unittest.TestCase):

    def test_type_mapping_error(self):
        # Test that returns an error if input is not one of single or multi
        with self.assertRaises(ValueError):
            Encoder(data_frame=make_df(), dataset_type='notsupportedtype')

    def test_type_mapping_single(self):
        enc = Encoder(data_frame=make_df(), dataset_type='single')
        self.assertEqual(enc._Encoder__orientation, 'records')

    def test_type_mapping_multi(self):
        enc = Encoder(data_frame=make_df(), dataset_type='multi')
        self.assertEqual(enc._Encoder__orientation, 'values')

    def test_single_table_encoding(self):
        # Test validity of encoding for single-table
        enc = Encoder(data_frame=make_df(), dataset_type='single')
        df = pd.DataFrame(json.loads(b64decode(enc._Encoder__b64_data)))
        self.assertEqual(df.salary.sum(), 560000)

    def test_multi_table_encoding(self):
        # Test validity of encoding of multi-table
        enc = Encoder(data_frame=make_df(), dataset_type='multi')
        df = pd.DataFrame(json.loads(b64decode(enc._Encoder__b64_data)),
                          columns=["id_int", "id_str", "first_name",
                                   "last_name", "age", "weight", "state", "salary"])
        self.assertEqual(df.salary.sum(), 560000)


if __name__ == '__main__':
    unittest.main()



import requests


def login(connection, verbose=False):
    """
    Authenticate a user and create an HTTP session on the web server where the user’s MicroStrategy sessions are stored.
    This request returns an authorization token (X-MSTR-AuthToken) which will be submitted with subsequent requests.
    The body of the request contains the information needed to create the session. The loginMode parameter in the body
    specifies the authentication mode to use. You can authenticate with one of the following authentication modes:
    Standard (1), Anonymous (8), or LDAP (16). Authentication modes can be enabled through the System Administration
    REST APIs, if they are supported by the deployment.

    :param connection: MicroStrategy REST API connection object
    :param verbose: Verbosity of request response; defaults to False
    :return: Complete HTTP response object
    """

    response = requests.post(url=connection.base_url + '/auth/login',
                             data={'username': connection.username,
                                   'password': connection.password,
                                   'loginMode': connection.login_mode,
                                   'applicationType': connection.application_code},
                             verify=connection.ssl_verify)
    if verbose:
        print(response.url)
    return response


def logout(connection, verbose=False):
    """
    Close all existing sessions for the authenticated user.

    :param connection: MicroStrategy REST API connection object
    :param verbose: Verbosity of request response; defaults to False
    :return: Complete HTTP response object
    """
    response = requests.post(url=connection.base_url + '/auth/logout',
                             headers={'X-MSTR-AuthToken': connection.auth_token},
                             cookies=connection.cookies,
                             verify=connection.ssl_verify)
    if verbose:
        print(response.url)
    return response


def sessions(connection, verbose=False):
    """
    Extends the HTTP and Intelligence Server sessions by resetting the timeouts.

    :param connection: MicroStrategy REST API connection object
    :param verbose: Verbosity of request response; defaults to False
    :return: Complete HTTP response object
    """
    response = requests.put(url=connection.base_url + '/sessions',
                            headers={'X-MSTR-AuthToken': connection.auth_token},
                            cookies=connection.cookies,
                            verify=connection.ssl_verify)
    if verbose:
        print(response.url)
    return response



import pandas as pd


def parsejson(response):
    """

    :param response:
    :return:
    """

    def _has_child(node):
        if 'children' in node.keys():
            return True
        else:
            return False

    def _get_attribute_names(_response):
        # Returns a list of attribute column names from the full json response object
        j = _response['result']['definition']['attributes']
        return [j[n]['name'] for n in range(len(j))]

    def _get_attribute(node):
        # Returns the attribute index and name from a node as a tuple
        return node['element']['attributeIndex'], node['element']['name']

    def _get_metric_names(_response):
        # Returns a list of metric names from the full json response object
        j = _response['result']['definition']['metrics']
        return [j[n]['name'] for n in range(len(j))]

    def _get_metrics(node):
        if 'metrics' in node.keys():
            _metrics = [node['metrics'][_key]['rv'] for _key in node['metrics']]
        else:
            _metrics = []  # in the case of a report with no metrics in it, return an empty list
        return _metrics

    def _parse_node(node):
        # Recursively parses the json string, extract attributes and metrics
        _data = []
        if _has_child(node=node):
            for child in node['children']:
                _data.append(_get_attribute(node=child))
                _data += _parse_node(node=child)
        else:
            _data.append(_get_metrics(node=node))
        return _data

    # if the response['result']['data']['root'] is None return None
    if response['result']['data']['root'] is None:
        return None

    attribute_names = _get_attribute_names(response)
    metric_names = _get_metric_names(response)
    parsed_rows = _parse_node(node=response['result']['data']['root'])

    metrs = []
    attrs = []
    temp = [None] * len(attribute_names)

    for element in parsed_rows:
        if isinstance(element, tuple):
            _attr_idx, _attr_name = element
            temp[_attr_idx] = _attr_name
            if _attr_idx == len(attribute_names)-1:
                attrs.append(temp[:])
        if isinstance(element, list):
            metrs.append(element)

    df_data = []
    for att, met in zip(attrs, metrs):
        df_data.append(att + met)

    # add metrics to df_data when attrs is empty and metrs is a single row of values
    if not attrs and metrs:
        [metric_values] = metrs
        df_data.append(metric_values)

    df = pd.DataFrame(df_data, columns=attribute_names + metric_names)

    return df
